import numpy as np
import librosa

from audiomate import tracks
from audiomate import annotations
from audiomate import issuers

import pytest
from tests import resources


class TestLabel:

    def test_label_creation(self):
        a = annotations.Label('value', 6.2, 8.9)

        assert a.value == 'value'
        assert a.start == 6.2
        assert a.end == 8.9
        assert len(a.meta) == 0

    def test_label_creation_with_info(self):
        a = annotations.Label('value', 6.2, 8.9, meta={'something': 2})

        assert a.value == 'value'
        assert a.start == 6.2
        assert a.end == 8.9
        assert len(a.meta) == 1
        assert a.meta['something'] == 2

    def test_lt_start_time_considered_first(self):
        a = annotations.Label('some label', 1.0, 2.0)
        b = annotations.Label('some label', 1.1, 2.0)

        assert a < b

    def test_lt_end_time_considered_second(self):
        a = annotations.Label('some label', 1.0, 1.9)
        b = annotations.Label('some label', 1.0, 2.0)

        assert a < b

    def test_lt_end_time_properly_handles_custom_infinity(self):
        a = annotations.Label('some label', 1.0, 1.9)
        b = annotations.Label('some label', 1.0, float('inf'))

        assert a < b

    def test_lt_value_considered_third(self):
        a = annotations.Label('some label a', 1.0, 2.0)
        b = annotations.Label('some label b', 1.0, 2.0)

        assert a < b

    def test_lt_value_ignores_capitalization(self):
        a = annotations.Label('some label A', 1.0, 2.0)
        b = annotations.Label('some label a', 1.0, 2.0)

        assert not a < b  # not == because == tests different method
        assert not a > b  # not == because == tests different method

    def test_eq_ignores_capitalization(self):
        a = annotations.Label('some label A', 1.0, 2.0)
        b = annotations.Label('some label a', 1.0, 2.0)

        assert a == b

    def test_eq_ignores_label_list_relation(self):
        a = annotations.Label('some label A', 1.0, 2.0)
        b = annotations.Label('some label a', 1.0, 2.0)

        al = annotations.LabelList(idx='one', labels=[a])
        bl = annotations.LabelList(idx='another', labels=[b])

        assert a.label_list == al
        assert b.label_list == bl
        assert a == b

    def test_lt_ignores_label_list_relation(self):
        a = annotations.Label('some label A', 1.0, 2.0)
        b = annotations.Label('some label a', 1.0, 2.0)

        al = annotations.LabelList(idx='one', labels=[a])
        bl = annotations.LabelList(idx='another', labels=[b])

        assert a.label_list == al
        assert b.label_list == bl
        assert not a < b
        assert not a > b

    def test_read_samples(self):
        path = resources.sample_wav_file('wav_1.wav')
        track = tracks.FileTrack('wav', path)
        issuer = issuers.Issuer('toni')
        utt = tracks.Utterance('t', track, issuer=issuer, start=1.0, end=2.30)

        l1 = annotations.Label('a', 0.15, 0.448)
        l2 = annotations.Label('a', 0.5, 0.73)
        ll = annotations.LabelList(labels=[l1, l2])

        utt.set_label_list(ll)

        expected, __ = librosa.core.load(
            path,
            sr=None,
            offset=1.15,
            duration=0.298
        )
        assert np.array_equal(l1.read_samples(), expected)

        expected, __ = librosa.core.load(
            path,
            sr=None,
            offset=1.5,
            duration=0.23
        )
        assert np.array_equal(l2.read_samples(), expected)

    def test_read_samples_no_utterance_and_label_end(self):
        path = resources.sample_wav_file('wav_1.wav')
        track = tracks.FileTrack('wav', path)
        issuer = issuers.Issuer('toni')
        utt = tracks.Utterance('idx', track, issuer=issuer, start=1.0)

        l1 = annotations.Label('a', 0.15, 0.448)
        l2 = annotations.Label('a', 0.5)
        ll = annotations.LabelList(labels=[l1, l2])

        utt.set_label_list(ll)

        expected, __ = librosa.core.load(
            path,
            sr=None,
            offset=1.15,
            duration=0.298
        )
        assert np.array_equal(l1.read_samples(), expected)

        expected, __ = librosa.core.load(path, sr=None, offset=1.5)
        assert np.array_equal(l2.read_samples(), expected)

    def test_start_abs(self):
        label = annotations.Label('a', 2, 5)
        ll = annotations.LabelList(labels=[label])
        tracks.Utterance('utt-1', None, start=1, end=19, label_lists=[ll])

        assert label.start_abs == 3

    def test_start_abs_no_utterance(self):
        label = annotations.Label('a', 2, 5)

        assert label.start_abs == 2

    def test_end_abs(self):
        label = annotations.Label('a', 2, 5)
        ll = annotations.LabelList(labels=[label])
        tracks.Utterance('utt-1', None, start=1, end=19, label_lists=[ll])

        assert label.end_abs == 6

    def test_end_abs_no_utterance(self):
        label = annotations.Label('a', 2, 5)

        assert label.end_abs == 5

    def test_duration(self):
        label = annotations.Label('a', 2, 5)

        assert label.duration == 3

    def test_length(self):
        label = annotations.Label('abcd', 2, 3)

        assert label.length == 4

    def test_tokenized(self):
        label = annotations.Label('wo wie was warum  weshalb')

        assert label.tokenized(delimiter=' ') == [
            'wo', 'wie', 'was', 'warum', 'weshalb'
        ]

    def test_tokenized_returns_empty_list_for_empty_label_value(self):
        label = annotations.Label('  ')

        assert label.tokenized(delimiter=' ') == []

    def test_tokenized_with_other_delimiter(self):
        label = annotations.Label('wie,kann, ich, dass machen')

        assert label.tokenized(delimiter=',') == ['wie', 'kann', 'ich', 'dass machen']

    def test_do_overlap_other_fully_contained(self):
        label = annotations.Label('a', 2.5, 9.5)
        label_other = annotations.Label('a', 4.8, 8.4)

        assert label.overlap_duration(label_other) == pytest.approx(8.4 - 4.8)

    def test_do_overlap_this_fully_contained(self):
        label = annotations.Label('a', 4.8, 8.4)
        label_other = annotations.Label('a', 2.5, 9.5)

        assert label.do_overlap(label_other)

    def test_do_overlap_left(self):
        label = annotations.Label('a', 4.8, 8.4)
        label_other = annotations.Label('a', 2.5, 5.3)

        assert label.do_overlap(label_other)

    def test_do_overlap_right(self):
        label = annotations.Label('a', 4.8, 8.4)
        label_other = annotations.Label('a', 7.5, 9.2)

        assert label.do_overlap(label_other)

    def test_do_overlap_none_right(self):
        label = annotations.Label('a', 4.8, 8.4)
        label_other = annotations.Label('a', 8.5, 9.2)

        assert not label.do_overlap(label_other)

    def test_do_overlap_none_left(self):
        label = annotations.Label('a', 4.8, 8.4)
        label_other = annotations.Label('a', 1.2, 3.2)

        assert not label.do_overlap(label_other)

    def test_do_overlap_none_self_infinite(self):
        label = annotations.Label('a', 4.8, float('inf'))
        label_other = annotations.Label('a', 1.2, 3.2)

        assert not label.do_overlap(label_other)

    def test_do_overlap_none_other_infinite(self):
        label = annotations.Label('a', 4.8, 10.5)
        label_other = annotations.Label('a', 10.6, float('inf'))

        assert not label.do_overlap(label_other)

    def test_do_overlap_this_infinite(self):
        label = annotations.Label('a', 4.8, float('inf'))
        label_other = annotations.Label('a', 8.5, 9.2)

        assert label.do_overlap(label_other)

    def test_do_overlap_other_infinite(self):
        label = annotations.Label('a', 4.8, 9.3)
        label_other = annotations.Label('a', 8.5, float('inf'))

        assert label.do_overlap(label_other)

    def test_do_overlap_both_infinite(self):
        label = annotations.Label('a', 4.8, float('inf'))
        label_other = annotations.Label('a', 8.5, float('inf'))

        assert label.do_overlap(label_other)

    def test_do_overlap_adjacent_allowed(self):
        label = annotations.Label('a', 4.8, 8.5)
        label_other = annotations.Label('a', 8.5, 9.3)

        assert label.do_overlap(label_other, adjacent=True)

    def test_do_overlap_adjacent_disallowed(self):
        label = annotations.Label('a', 4.8, 8.5)
        label_other = annotations.Label('a', 8.5, 9.3)

        assert not label.do_overlap(label_other, adjacent=False)

    def test_overlap_duration_other_fully_contained(self):
        label = annotations.Label('a', 2.5, 9.5)
        label_other = annotations.Label('a', 4.8, 8.4)

        assert label.overlap_duration(label_other) == pytest.approx(8.4 - 4.8)

    def test_overlap_duration_this_fully_contained(self):
        label = annotations.Label('a', 4.8, 8.4)
        label_other = annotations.Label('a', 2.5, 9.5)

        assert label.overlap_duration(label_other) == pytest.approx(8.4 - 4.8)

    def test_overlap_duration_left(self):
        label = annotations.Label('a', 4.8, 8.4)
        label_other = annotations.Label('a', 2.5, 5.3)

        assert label.overlap_duration(label_other) == pytest.approx(5.3 - 4.8)

    def test_overlap_duration_right(self):
        label = annotations.Label('a', 4.8, 8.4)
        label_other = annotations.Label('a', 7.5, 9.2)

        assert label.overlap_duration(label_other) == pytest.approx(8.4 - 7.5)

    def test_overlap_duration_none(self):
        label = annotations.Label('a', 4.8, 8.4)
        label_other = annotations.Label('a', 8.5, 9.2)

        assert label.overlap_duration(label_other) == 0.0

    def test_overlap_duration_this_infinite(self):
        label = annotations.Label('a', 4.8, float('inf'))
        label_other = annotations.Label('a', 8.5, 9.2)

        assert label.overlap_duration(label_other) == pytest.approx(9.2 - 8.5)

    def test_overlap_duration_other_infinite(self):
        label = annotations.Label('a', 4.8, 9.3)
        label_other = annotations.Label('a', 8.5, float('inf'))

        assert label.overlap_duration(label_other) == pytest.approx(9.3 - 8.5)

    def test_overlap_duration_both_infinite(self):
        label = annotations.Label('a', 4.8, float('inf'))
        label_other = annotations.Label('a', 8.5, float('inf'))

        assert label.overlap_duration(label_other) == float('inf')



"""
This module contains the different implementations of a track.
A track is an abstract representation of an audio signal.

A concrete implementation provides the functionalty
for reading the audio samples from a specific source.
"""

from .track import Track  # noqa: F401
from .file import FileTrack  # noqa: F401
from .container import ContainerTrack  # noqa: F401

from .utterance import Utterance  # noqa: F401



import bisect
import math
import random

import numpy as np

import audiomate
from audiomate import containers
from . import partitioning


class DataIterator(object):
    """
    An abstract class representing a data-iterator. A data-iterator provides sequential access to data.
    An implementation of a concrete data-iterator should override the methods ``__iter__`` and ``__next__``.

    A sample returned from a data-iterator is a tuple containing the data for this sample from every container.
    The data from different containers is ordered in the way the containers were passed to the DataIterator.

    Args:
        corpus_or_utt_ids (Corpus, list): Either a corpus or a list of utterances.
                                          This defines which utterances are considered for iterating.
        containers (list, Container): A single container or a list of containers.
        shuffle (bool): Indicates whether the data should be returned in
                        random order (``True``) or not (``False``).
        seed (int): Seed to be used for the random number generator.
    """

    def __init__(self, corpus_or_utt_ids, feature_containers, shuffle=True, seed=None):
        if isinstance(corpus_or_utt_ids, audiomate.corpus.CorpusView):
            self.utt_ids = list(corpus_or_utt_ids.utterances.keys())
        else:
            self.utt_ids = corpus_or_utt_ids

        if isinstance(feature_containers, containers.Container):
            self.containers = [feature_containers]
        else:
            self.containers = feature_containers

        if len(self.containers) == 0:
            raise ValueError('At least one container has to be provided!')

        self.shuffle = shuffle

        self.rand = random.Random()
        self.rand.seed(a=seed)

    def __iter__(self):
        raise NotImplementedError

    def __next__(self):
        raise NotImplementedError


class MultiFrameIterator(DataIterator):
    """
    A data-iterator wrapping chunks of subsequent frames of a corpus.
    A single sample represents a chunk of frames.

    Args:
        corpus_or_utt_ids (Corpus, list): Either a corpus or a list of utterances.
                                          This defines which utterances are considered for iterating.
        container (list, Container): A single container or a list of containers.
        partition_size (str): Size of the partitions in bytes. The units ``k`` (kibibytes), ``m`` (mebibytes) and ``g``
                              (gibibytes) are supported, i.e. a ``partition_size`` of ``1g`` equates :math:`2^{30}`
                              bytes.
        frames_per_chunk (int): Number of subsequent frames in a single sample.
        return_length (bool): If True, the length of the chunk is returned as well. (default ``False``)
                              The length is appended to tuple as the last element.
                              (e.g. [container1-data, container2-data, length])
        pad (bool): If True, samples that are shorter are padded with zeros to match ``frames_per_chunk``.
                    If padding is enabled, the lengths are always returned ``return_length = True``.
        shuffle (bool): Indicates whether the data should be returned in
                        random order (``True``) or not (``False``).
        seed (int): Seed to be used for the random number generator.

    Note:
        For a MultiFrameIterator it is expected that every container contains exactly one value/vector for every frame.
        So the first dimension (outermost) of every array in every container have to match.

    Example:
        >>> corpus = audiomate.Corpus.load('/path/to/corpus')
        >>> container_inputs = containers.FeatureContainer('/path/to/features.hdf5')
        >>> container_outputs = containers.Container('/path/to/targets.hdf5')
        >>>
        >>> ds = MultiFrameIterator(corpus, [container_inputs, container_outputs], '1G', 5, shuffle=True, seed=23)
        >>> next(ds) # Next Chunk (inputs, outputs)
        (
            array([[0.72991909, 0.20258683, 0.30574747, 0.53783217],
                   [0.38875413, 0.83611128, 0.49054591, 0.15710017],
                   [0.35153358, 0.40051009, 0.93647765, 0.29589257],
                   [0.97465772, 0.80160451, 0.81871436, 0.4892925 ],
                   [0.59310933, 0.8565602 , 0.95468696, 0.07933512]])
            array([[0.0, 1.0], [0.0, 1.0],[0.0, 1.0],[0.0, 1.0], [0.0, 1.0]])
        )
    """

    def __init__(self, corpus_or_utt_ids, containers, partition_size, frames_per_chunk, return_length=False,
                 pad=False, shuffle=True, seed=None):
        super(MultiFrameIterator, self).__init__(corpus_or_utt_ids, containers, shuffle=shuffle, seed=seed)

        self.partition_size = partition_size
        self.frames_per_chunk = frames_per_chunk
        self.pad = pad

        if self.pad:
            self.return_length = True
        else:
            self.return_length = return_length

        self.loader = None
        self.current_partition = None
        self.current_partition_index = -1
        self.current_chunk_index = 0

        self.loader = partitioning.PartitioningContainerLoader(self.utt_ids,
                                                               self.containers,
                                                               self.partition_size,
                                                               shuffle=self.shuffle,
                                                               seed=self.rand.random())

    def __iter__(self):
        self.current_partition = None
        self.current_partition_index = -1
        self.current_chunk_index = 0

        self.loader.reload()

        return self

    def __next__(self):
        if self.current_partition is None or self.current_chunk_index >= len(self.current_partition):
            self.current_partition_index += 1
            self.current_chunk_index = 0

            if self.current_partition_index < len(self.loader.partitions):
                partition_data = self.loader.load_partition_data(self.current_partition_index)
                self.current_partition = MultiFramePartitionData(partition_data,
                                                                 self.frames_per_chunk,
                                                                 return_length=self.return_length,
                                                                 pad=self.pad,
                                                                 shuffle=self.shuffle,
                                                                 seed=self.rand.random())
            else:
                raise StopIteration

        next_chunk = self.current_partition[self.current_chunk_index]
        self.current_chunk_index += 1

        return next_chunk


class FrameIterator(MultiFrameIterator):
    """
    A data-iterator wrapping frames of a corpus. A single sample represents a single frame.

    Args:
        corpus_or_utt_ids (Corpus, list): Either a corpus or a list of utterances.
                                          This defines which utterances are considered for iterating.
        container (list, Container): A single container or a list of containers.
        partition_size (str): Size of the partitions in bytes. The units ``k`` (kibibytes), ``m`` (mebibytes) and ``g``
                              (gibibytes) are supported, i.e. a ``partition_size`` of ``1g`` equates :math:`2^{30}`
                              bytes.
        shuffle (bool): Indicates whether the data should be returned in
                        random order (``True``) or not (``False``).
        seed (int): Seed to be used for the random number generator.

    Note:
        For a FrameIterator it is expected that every container contains exactly one value/vector for every frame.
        So the first dimension of every array in every container have to match.

    Example:
        >>> corpus = audiomate.Corpus.load('/path/to/corpus')
        >>> container_inputs = containers.FeatureContainer('/path/to/features.hdf5')
        >>> container_outputs = containers.Container('/path/to/targets.hdf5')
        >>>
        >>> ds = FrameIterator(corpus, [container_inputs, container_outputs], '1G', shuffle=True, seed=23)
        >>> next(ds) # Next Frame (inputs, outputs)
        (
            array([0.58843831, 0.18128443, 0.19718328, 0.25284105]),
            array([0.0, 1.0])
        )
    """

    def __init__(self, corpus_or_utt_ids, containers, partition_size, shuffle=True, seed=None):
        super(FrameIterator, self).__init__(corpus_or_utt_ids, containers, partition_size, 1,
                                            return_length=False, shuffle=shuffle, seed=seed)

    def __next__(self):
        data = super(FrameIterator, self).__next__()

        # We have to remove the outermost dimension, which is 1 for chunk-size of 1 frame
        return [x[0] for x in data]


class MultiFramePartitionData(object):
    """
    Wrapper for PartitionData to access chunks of frames via indexes.

    Args:
        partition_data (PartitionData): The loaded partition-data.
        frames_per_chunk (int): Number of subsequent frames in a chunk.
        return_length (bool): If True, the length of the chunk is returned as well. (default ``False``)
                              The length is appended to tuple as the last element.
                              (e.g. [container1-data, container2-data, length])
        pad (bool): If True, samples that are shorter are padded with zeros to match ``frames_per_chunk``.
                    If padding is enabled, the lengths are always returned ``return_length = True``.
        shuffle (bool): If True the frames are shuffled randomly for access.
        seed (int): The seed to use for shuffling.
    """

    def __init__(self, partition_data, frames_per_chunk, return_length=False, pad=False, shuffle=True, seed=None):
        if frames_per_chunk < 1:
            raise ValueError('Number of frames per chunk has to higher than 0.')

        self.data = partition_data
        self.frames_per_chunk = frames_per_chunk
        self.pad = pad

        if self.pad:
            self.return_length = True
        else:
            self.return_length = return_length

        self.shuffle = shuffle

        self.rand = random.Random()
        self.rand.seed(a=seed)

        # Regions are used to provide indexed access across all utterances
        self.regions = self.get_utt_regions()
        self.region_offsets = [x[0] for x in self.regions]

        # Sampling used to access frames
        self.sampling = list(range(len(self)))

        if self.shuffle:
            self.rand.shuffle(self.sampling)

    def __len__(self):
        last_region = self.regions[-1]
        return last_region[0] + last_region[1]

    def __getitem__(self, item):
        index = self.sampling[item]

        # we search the region before the offset is higher than the index.
        region_index = bisect.bisect_right(self.region_offsets, index) - 1
        region = self.regions[region_index]

        frame_offset = (index - region[0]) * self.frames_per_chunk
        frame_end = frame_offset + self.frames_per_chunk

        data = [x[frame_offset:frame_end].astype(np.float32) for x in region[2]]
        size = data[0].shape[0]

        if self.pad and size < self.frames_per_chunk:
            padded_data = []
            for x in data:
                # Only pad the outermost (first) dimension
                pad_widths = [(0, 0)] * (len(x.shape) - 1)
                pad_widths.insert(0, (0, self.frames_per_chunk - size))
                padded_x = np.pad(x, pad_widths, mode='constant', constant_values=0)
                padded_data.append(padded_x)

            data = padded_data

        if self.return_length:
            data.append(size)

        return data

    def get_utt_regions(self):
        """
        Return the regions of all utterances, assuming all utterances are concatenated.
        A region is defined by offset, length (num-frames) and
        a list of references to the utterance datasets in the containers.

        Returns:
            list: List of with a tuple for every utterances containing the region info.
        """

        regions = []
        current_offset = 0

        for utt_idx, utt_data in zip(self.data.info.utt_ids, self.data.utt_data):
            offset = current_offset

            num_frames = []
            refs = []

            for part in utt_data:
                num_frames.append(part.shape[0])
                refs.append(part)

            if len(set(num_frames)) != 1:
                raise ValueError('Utterance {} has not the same number of frames in all containers!'.format(utt_idx))

            num_chunks = math.ceil(num_frames[0] / float(self.frames_per_chunk))

            region = (offset, num_chunks, refs)
            regions.append(region)

            # Sets the offset for the next utterances
            current_offset += num_chunks

        return regions



import statistics
import random

from audiomate.corpus.subset import utils


def test_absolute_proportions():
    res = utils.absolute_proportions({
        'a': 0.6,
        'b': 0.2,
        'c': 0.2
    }, 120)

    assert res['a'] == 72
    assert res['b'] == 24
    assert res['c'] == 24


def test_split_identifiers():
    identifiers = [str(x) for x in range(10)]
    random.shuffle(identifiers)
    proportions = {'a': 0.3, 'b': 0.4, 'c': 0.3}

    result = utils.split_identifiers(identifiers, proportions, seed=220)

    assert len(result['a']) == 3
    assert sorted(result['a']) == ['0', '1', '6']
    assert len(result['b']) == 4
    assert sorted(result['b']) == ['2', '4', '5', '8']
    assert len(result['c']) == 3
    assert sorted(result['c']) == ['3', '7', '9']


def test_get_identifiers_randomly_splitted():
    res = utils.split_identifiers(identifiers=[
        'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'v', 't'
    ], proportions={
        'a': 0.3333,
        'b': 0.6666
    })

    assert len(res['a']) == 4
    assert len(res['b']) == 8
    assert len(set(res['a'] + res['b'])) == 12


def test_get_identifiers_splitted_by_weights_single_category():
    identifiers = {
        'a': {'mi': 3},
        'b': {'mi': 4},
        'c': {'mi': 6},
        'd': {'mi': 1},
        'e': {'mi': 4},
        'f': {'mi': 5},
        'g': {'mi': 3}
    }

    proportions = {
        'train': 0.5,
        'test': 0.25,
        'dev': 0.25
    }

    res = utils.get_identifiers_splitted_by_weights(identifiers=identifiers,
                                                    proportions=proportions)

    assert len(res['train']) > 0
    assert len(res['test']) > 0
    assert len(res['dev']) > 0
    assert len(identifiers) == sum([len(x) for x in res.values()])


def test_get_identifiers_splitted_by_weights():
    identifiers = {
        'a': {'mi': 3, 'ma': 2, 'mu': 1},
        'b': {'mi': 4, 'ma': 5, 'mu': 4},
        'c': {'mi': 6, 'ma': 4, 'mu': 3},
        'd': {'mi': 1, 'ma': 3, 'mu': 2},
        'e': {'mi': 4, 'ma': 1, 'mu': 5},
        'f': {'mi': 5, 'ma': 4, 'mu': 3},
        'g': {'mi': 3, 'ma': 4, 'mu': 5}
    }

    proportions = {
        'train': 0.5,
        'test': 0.25,
        'dev': 0.25
    }

    res = utils.get_identifiers_splitted_by_weights(identifiers=identifiers,
                                                    proportions=proportions)

    assert len(res['train']) > 0
    assert len(res['test']) > 0
    assert len(res['dev']) > 0
    assert len(identifiers) == sum([len(x) for x in res.values()])


def test_get_identifiers_splitted_by_weights_randomly():
    identifiers = {
        'a': {'mi': 3, 'ma': 2, 'mu': 1},
        'b': {'mi': 4, 'ma': 5, 'mu': 4},
        'c': {'mi': 6, 'ma': 4, 'mu': 3},
        'd': {'mi': 1, 'ma': 3, 'mu': 2},
        'e': {'mi': 4, 'ma': 1, 'mu': 5},
        'f': {'mi': 5, 'ma': 4, 'mu': 3},
        'g': {'mi': 3, 'ma': 4, 'mu': 5}
    }

    proportions = {
        'train': 0.5,
        'test': 0.25,
        'dev': 0.25
    }

    res = utils.get_identifiers_splitted_by_weights(
        identifiers=identifiers,
        proportions=proportions,
        seed=220
    )

    assert len(res['train']) == 3
    assert res['train'] == ['d', 'g', 'b']
    assert len(res['test']) == 2
    assert res['test'] == ['c', 'f']
    assert len(res['dev']) == 2
    assert res['dev'] == ['a', 'e']


def test_get_identifiers_splitted_by_weights_assumes_zero_if_no_cat_weight():
    identifiers = {
        'a': {'mi': 6, 'ma': 2, 'mu': 1},
        'b': {'mi': 4, 'ma': 5, 'mu': 4},
        'c': {'ma': 4, 'mu': 3},
        'd': {'mi': 1, 'ma': 3},
        'e': {'mi': 4, 'ma': 1, 'mu': 5},
        'f': {'mi': 5, 'ma': 4, 'mu': 3},
        'g': {'mi': 3, 'ma': 4, 'mu': 5}
    }

    proportions = {
        'train': 0.5,
        'test': 0.25,
        'dev': 0.25
    }

    res = utils.get_identifiers_splitted_by_weights(
        identifiers=identifiers,
        proportions=proportions,
        seed=220
    )

    assert len(res['train']) == 3
    assert res['train'] == ['g', 'b', 'f']
    assert len(res['test']) == 2
    assert res['test'] == ['e', 'd']
    assert len(res['dev']) == 2
    assert res['dev'] == ['a', 'c']


def test_select_balanced_subset():
    categories = ['a', 'b', 'c']
    items = {
        'utt-1': {'a': 1, 'b': 0, 'c': 0},
        'utt-2': {'a': 1, 'c': 2},
        'utt-3': {'a': 1, 'b': 1, 'c': 1},
        'utt-4': {'a': 1, 'b': 1, 'c': 0},
        'utt-5': {'b': 1, 'c': 0},
        'utt-6': {'b': 2, 'c': 1},
        'utt-7': {'a': 1, 'b': 0, 'c': 1},
        'utt-8': {'a': 1, 'b': 1, 'c': 0},
        'utt-9': {'a': 1, 'b': 0, 'c': 0},
        'utt-10': {'c': 2},
        'utt-11': {'a': 1, 'b': 1, 'c': 1},
        'utt-12': {'b': 1, 'c': 0},
        'utt-13': {'b': 1, 'c': 0},
        'utt-14': {'b': 2, 'c': 1},
        'utt-15': {'a': 1, 'b': 0, 'c': 1},
        'utt-16': {'a': 1, 'b': 1, 'c': 0}
    }

    item_ids = utils.select_balanced_subset(items, 10, categories, seed=33)
    weights = [0 for __ in categories]

    for item_id in item_ids:
        for cat, weight in items[item_id].items():
            weights[categories.index(cat)] += weight

    assert len(item_ids) == 10
    assert statistics.variance(weights) <= 1



import os

import pytest
import requests_mock

from audiomate import corpus
from audiomate.corpus import io
from audiomate.corpus.io import rouen

from tests import resources
from . import reader_test as rt


@pytest.fixture
def downloader():
    return io.RouenDownloader()


@pytest.fixture()
def zip_data():
    with open(resources.get_resource_path(['sample_files', 'zip_sample_with_subfolder.zip']), 'rb') as f:
        return f.read()


class TestRouenDownloader:

    def test_download(self, zip_data, downloader, tmpdir):
        target_folder = tmpdir.strpath

        with requests_mock.Mocker() as mock:
            mock.get(rouen.DATA_URL, content=zip_data)

            downloader.download(target_folder)

        assert os.path.isfile(os.path.join(target_folder, 'a.txt'))
        assert os.path.isfile(os.path.join(target_folder, 'subsub', 'b.txt'))
        assert os.path.isfile(os.path.join(target_folder, 'subsub', 'c.txt'))


class TestRouenReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('rouen')
    FILE_TRACK_BASE_PATH = SAMPLE_PATH

    EXPECTED_NUMBER_OF_TRACKS = 4
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('avion1', 'avion1.wav'),
        rt.ExpFileTrack('avion2', 'avion2.wav'),
        rt.ExpFileTrack('bus1', 'bus1.wav'),
        rt.ExpFileTrack('metro_rouen22', 'metro_rouen22.wav'),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 0

    EXPECTED_NUMBER_OF_UTTERANCES = 4
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('avion1', 'avion1', None, 0, float('inf')),
        rt.ExpUtterance('avion2', 'avion2', None, 0, float('inf')),
        rt.ExpUtterance('bus1', 'bus1', None, 0, float('inf')),
        rt.ExpUtterance('metro_rouen22', 'metro_rouen22', None, 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        'avion1': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
        'avion2': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
        'bus1': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
        'metro_rouen22': [rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1)],
    }

    EXPECTED_LABELS = {
        'avion1':   [rt.ExpLabel(corpus.LL_SOUND_CLASS, 'avion', 0, float('inf'))],
        'avion2': [rt.ExpLabel(corpus.LL_SOUND_CLASS, 'avion', 0, float('inf'))],
        'bus1':   [rt.ExpLabel(corpus.LL_SOUND_CLASS, 'bus', 0, float('inf'))],
        'metro_rouen22':   [rt.ExpLabel(corpus.LL_SOUND_CLASS, 'metro_rouen', 0, float('inf'))],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 0

    def load(self):
        return io.RouenReader().load(self.SAMPLE_PATH)



import abc

import numpy as np
import networkx as nx

from audiomate import processing


class Chunk(object):
    """
    Represents a chunk of data. It is used to pass data between different steps of a pipeline.

    Args:
        data (np.ndarray or list): A single array of frames or a list of separate chunks of frames of equal size.
        offset (int): The index of the first frame in the chunk within the sequence.
        is_last (bool): Whether this is the last chunk of the sequence.
        left_context (int): Number of frames that act as context at the begin of the chunk (left).
        right_context (int): Number of frames that act as context at the end of the chunk (right).
    """

    def __init__(self, data, offset, is_last, left_context=0, right_context=0):
        self.data = data
        self.offset = offset
        self.is_last = is_last
        self.left_context = left_context
        self.right_context = right_context

    def __repr__(self):
        return 'Chunk(data [{}], offset [{}], is-last [{}], left[{}], right[{}])'.format(self.data.shape,
                                                                                         self.offset,
                                                                                         self.is_last,
                                                                                         self.left_context,
                                                                                         self.right_context)


class Buffer(object):
    """
    The buffer is a utility to store frames if there are not enough frames as required by some pipeline step.
    The incoming frames are added to the buffer using the ``update`` method.
    The buffer then determines if there are enough frames for a new chunk. If so a chunk is returned,
    otherwise the data is appended to the buffer.

    Context at the beginning or end of the sequence is not padded and is indicated by the chunks context attributes,
    which for example will be 0 for the left context in the first chunk.

    Using ``num_buffers`` multiple parallel buffers can be used. This means that a chunk is only returned,
    if all buffers have enough frames.

    Args:
        min_frames (int): The minimal number of frames needed in one chunk.
        left_context (int): The number of frames required as left context.
        right_context (int): The number of frames required as right context.
        num_buffers (int): The number of parallel buffers.

    Example:

        >>> b = Buffer(3, 2, 4)
        >>> b.update(np.arange(20).reshape(10,2), 0, False)
        Chunk(data [(10, 2)], offset [0], is-last [False], left[0], right[4])

        array([[ 0,  1],
               [ 2,  3],
               [ 4,  5],
               [ 6,  7],
               [ 8,  9],
               [10, 11],
               [12, 13],
               [14, 15],
               [16, 17],
               [18, 19]])
        >>> b.update(np.arange(2).reshape(1,2), 10, False)
        None
        >>> b.update(np.arange(6).reshape(3,2), 11, False)
        Chunk(data [(10, 2)], offset [4], is-last [False], left[2], right[4])

        array([[ 8,  9],
               [10, 11],
               [12, 13],
               [14, 15],
               [16, 17],
               [18, 19],
               [ 0,  1],
               [ 0,  1],
               [ 2,  3],
               [ 4,  5]])
        >>> b.update(np.arange(2).reshape(1,2), 14, True)
        Chunk(data [(7, 2)], offset [8], is-last [True], left[2], right[0])

        array([[16, 17],
               [18, 19],
               [ 0,  1],
               [ 0,  1],
               [ 2,  3],
               [ 4,  5],
               [ 0,  1]])
    """

    def __init__(self, min_frames, left_context, right_context, num_buffers=1):
        if num_buffers < 1:
            raise ValueError('Number of buffers has to be a positive int (>= 1).')

        self.min_frames = min_frames
        self.left_context = left_context
        self.right_context = right_context
        self.num_buffers = num_buffers

        self.buffers = [None] * self.num_buffers
        self.buffers_full = [False] * self.num_buffers

        self.current_frame = 0
        self.current_left_context = 0

    def update(self, data, offset, is_last, buffer_index=0):
        """
        Update the buffer at the given index.

        Args:
            data (np.ndarray): The frames.
            offset (int): The index of the first frame in `data` within the sequence.
            is_last (bool): Whether this is the last block of frames in the sequence.
            buffer_index (int): The index of the buffer to update (< self.num_buffers).
        """
        if buffer_index >= self.num_buffers:
            raise ValueError('Expected buffer index < {} but got index {}.'.format(self.num_buffers, buffer_index))

        if self.buffers[buffer_index] is not None and self.buffers[buffer_index].shape[0] > 0:
            expected_next_frame = self.current_frame + self.buffers[buffer_index].shape[0]
            if expected_next_frame != offset:
                raise ValueError(
                    'There are missing frames. Last frame in buffer is {}. The passed frames start at {}.'.format(
                        expected_next_frame, offset))

            self.buffers[buffer_index] = np.vstack([self.buffers[buffer_index], data])
        else:
            self.buffers[buffer_index] = data

        self.buffers_full[buffer_index] = is_last

    def get(self):
        """
        Get a new chunk if available.

        Returns:
            Chunk or list: If enough frames are available a chunk is returned. Otherwise None.
                           If ``self.num_buffer >= 1`` a list instead of single chunk is returned.
        """
        chunk_size = self._smallest_buffer()
        all_full = self._all_full()

        if all_full:
            right_context = 0
            num_frames = chunk_size - self.current_left_context
        else:
            right_context = self.right_context
            num_frames = self.min_frames

        chunk_size_needed = num_frames + self.current_left_context + right_context

        if chunk_size >= chunk_size_needed:
            data = []
            keep_frames = self.left_context + self.right_context
            keep_from = max(0, chunk_size - keep_frames)

            for index in range(self.num_buffers):
                data.append(self.buffers[index][:chunk_size])
                self.buffers[index] = self.buffers[index][keep_from:]

            if self.num_buffers == 1:
                data = data[0]

            chunk = Chunk(data,
                          self.current_frame,
                          all_full,
                          self.current_left_context,
                          right_context)

            self.current_left_context = min(self.left_context, chunk_size)
            self.current_frame = max(self.current_frame + chunk_size - keep_frames, 0)

            return chunk

    def _smallest_buffer(self):
        """
        Get the size of the smallest buffer.
        """

        smallest = np.inf

        for buffer in self.buffers:
            if buffer is None:
                return 0
            elif buffer.shape[0] < smallest:
                smallest = buffer.shape[0]

        return smallest

    def _all_full(self):
        """
        Return True if all buffers are full (last frame added).
        """
        for is_full in self.buffers_full:
            if not is_full:
                return False

        return True


class Step(processing.Processor, metaclass=abc.ABCMeta):
    """
    This class is the base class for a step in a processing pipeline.

    It handles the procedure of executing the pipeline. It makes sure the steps are computed in the correct order.
    It also provides the correct inputs to every step.

    Every step has to provide a ``compute`` method which is the actual processing.

    If the implementation of a step does change the frame or hop-size,
    it is expected to provide a transform via the ``frame_transform_step`` method.
    Frame-size and hop-size are measured in samples regarding the original audio signal (or simply its sampling rate).

    Args:
        name (str, optional): A name for identifying the step.
    """

    def __init__(self, name=None, min_frames=1, left_context=0, right_context=0):
        self.graph = nx.DiGraph()
        self.name = name

        self.min_frames = min_frames
        self.left_context = left_context
        self.right_context = right_context

        self.steps_sorted = []
        self.buffers = {}
        self.target_buffers = {}

    def process_frames(self, data, sampling_rate, offset=0, last=False, utterance=None, corpus=None):
        """
        Execute the processing of this step and all dependent parent steps.
        """

        if offset == 0:
            self.steps_sorted = list(nx.algorithms.dag.topological_sort(self.graph))
            self._create_buffers()
            self._define_output_buffers()

        # Update buffers with input data
        self._update_buffers(None, data, offset, last)

        # Go through the ordered (by dependencies) steps
        for step in self.steps_sorted:

            chunk = self.buffers[step].get()

            if chunk is not None:
                res = step.compute(chunk, sampling_rate, utterance=utterance, corpus=corpus)

                # If step is self, we know its the last step so return the data
                if step == self:
                    return res

                # Otherwise update buffers of child steps
                else:
                    self._update_buffers(step, res, chunk.offset + chunk.left_context, chunk.is_last)

    def frame_transform(self, frame_size, hop_size):
        parent_steps = self._parent_steps(self)

        if len(parent_steps) > 0:
            parent_frame_size, parent_hop_size = parent_steps[0].frame_transform(frame_size, hop_size)

            if len(parent_steps) > 1:
                # If there are multiple parents, we ensure that the frame-size is equal from all parents
                for i in range(1, len(parent_steps)):
                    next_fs, next_hs = parent_steps[i].frame_transform(frame_size, hop_size)

                    if next_fs != parent_frame_size:
                        raise ValueError('Frame-size from different differs!')

                    if next_hs != parent_hop_size:
                        raise ValueError('Hop-size from different differs!')

            return self.frame_transform_step(parent_frame_size, parent_hop_size)

        return self.frame_transform_step(frame_size, hop_size)

    @abc.abstractmethod
    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        """
        Do the computation of the step. If the step uses context, the result has to be returned without context.

        Args:
            chunk (Chunk): The chunk containing data and info about context, offset, ...
            sampling_rate (int): The sampling rate of the underlying signal.
            corpus (Corpus): The corpus the data is from, if available.
            utterance (Utterance): The utterance the data is from, if available.

        Returns:
            np.ndarray: The array of processed frames, without context.
        """
        pass

    def frame_transform_step(self, frame_size, hop_size):
        """
        If the processor changes the number of samples that build up a frame or
        the number of samples between two consecutive frames (hop-size),
        this function needs transform the original frame- and/or hop-size.

        This is used to store the frame-size and hop-size in a feature-container.
        In the end one can calculate start and end time of a frame with this information.

        By default it is assumed that the processor doesn't change the frame-size and the hop-size.

        Note:
            This function is simply for this step, whereas ``frame_transform()``
            computes the transformation for the whole pipeline.

        Args:
            frame_size (int): The original frame-size.
            hop_size (int): The original hop-size.

        Returns:
            tuple: The (frame-size, hop-size) after processing.
        """
        return frame_size, hop_size

    def _update_buffers(self, from_step, data, offset, is_last):
        """
        Update the buffers of all steps that need data from ``from_step``.
        If ``from_step`` is None it means the data is the input data.
        """

        for to_step, buffer in self.target_buffers[from_step]:
            parent_index = 0

            # if there multiple inputs we have to get the correct index, to keep the ordering
            if isinstance(to_step, Reduction):
                parent_index = to_step.parents.index(from_step)

            buffer.update(data, offset, is_last, buffer_index=parent_index)

    def _define_output_buffers(self):
        """
        Prepare a dictionary so we know what buffers have to be update with the the output of every step.
        """

        # First define buffers that need input data
        self.target_buffers = {
            None: [(step, self.buffers[step]) for step in self._get_input_steps()]
        }

        # Go through all steps and append the buffers of their child nodes
        for step in self.steps_sorted:
            if step != self:
                child_steps = [edge[1] for edge in self.graph.out_edges(step)]
                self.target_buffers[step] = [(child_step, self.buffers[child_step]) for child_step in child_steps]

    def _get_input_steps(self):
        """
        Search and return all steps that have no parents. These are the steps that are get the input data.
        """
        input_steps = []

        for step in self.steps_sorted:
            parent_steps = self._parent_steps(step)

            if len(parent_steps) == 0:
                input_steps.append(step)

        return input_steps

    def _create_buffers(self):
        """
        Create a buffer for every step in the pipeline.
        """

        self.buffers = {}

        for step in self.graph.nodes():
            num_buffers = 1

            if isinstance(step, Reduction):
                num_buffers = len(step.parents)

            self.buffers[step] = Buffer(step.min_frames, step.left_context, step.right_context, num_buffers)

        return self.buffers

    def _parent_steps(self, step):
        """ Return a list of all parent steps. """
        return [edge[0] for edge in self.graph.in_edges(step)]


class Computation(Step, metaclass=abc.ABCMeta):
    """
    Base class for a computation step.
    To implement a computation step for pipeline the ``compute`` method has to be implemented.
    This method gets the frames from its parent step including context frames if defined.
    It has to return the same number of frames but without context frames.

    Args:
        parent (Step, optional): The parent step this step depends on.
        name (str, optional): A name for identifying the step.
    """

    def __init__(self, parent=None, name=None, min_frames=1, left_context=0, right_context=0):
        super(Computation, self).__init__(name=name, min_frames=min_frames,
                                          left_context=left_context, right_context=right_context)

        self.graph.add_node(self)

        if parent is not None:
            self.graph.add_nodes_from(parent.graph.nodes())
            self.graph.add_edges_from(parent.graph.edges())
            self.graph.add_edge(parent, self)

    def __repr__(self) -> str:
        if self.name is None:
            return 'Computation'
        else:
            return self.name


class Reduction(Step, metaclass=abc.ABCMeta):
    """
    Base class for a reduction step.
    It gets the frames of all its parent steps as a list.
    It has to return a single chunk of frames.

    Args:
        parents (list): List of parent steps this step depends on.
        name (str, optional): A name for identifying the step.
    """

    def __init__(self, parents, name=None, min_frames=1, left_context=0, right_context=0):
        super(Reduction, self).__init__(name=name, min_frames=min_frames,
                                        left_context=left_context, right_context=right_context)

        self.parents = list(parents)
        self.graph.add_node(self)

        for index, parent in enumerate(parents):
            self.graph.add_nodes_from(parent.graph.nodes())
            self.graph.add_edges_from(parent.graph.edges())
            self.graph.add_edge(parent, self)

    def __repr__(self) -> str:
        if self.name is None:
            return 'Reduction'
        else:
            return self.name



import os

import pytest

from audiomate import issuers
from audiomate.corpus import io
from audiomate.utils import jsonfile

from tests import resources
from . import reader_test as rt


@pytest.fixture()
def writer():
    return io.DefaultWriter()


@pytest.fixture()
def sample_corpus():
    return resources.create_dataset()


class TestDefaultReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('default')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'files')

    EXPECTED_NUMBER_OF_TRACKS = 6
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('file-1', 'wav_1.wav'),
        rt.ExpFileTrack('file-2', 'wav_2.wav'),
        rt.ExpFileTrack('file-3', 'wav_3.wav'),
        rt.ExpFileTrack('file-4', 'wav_4.wav'),
        rt.ExpContainerTrack('file-5', 'file-5', os.path.join(SAMPLE_PATH, 'audio')),
        rt.ExpContainerTrack('file-6', 'file-6', os.path.join(SAMPLE_PATH, 'audio')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 3
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('speaker-1', 2, issuers.Gender.MALE, issuers.AgeGroup.ADULT, 'deu'),
        rt.ExpArtist('speaker-2', 2, 'Ohooo'),
        rt.ExpIssuer('speaker-3', 1),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 5
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('utt-1', 'file-1', 'speaker-1', 0, float('inf')),
        rt.ExpUtterance('utt-2', 'file-2', 'speaker-1', 0, float('inf')),
        rt.ExpUtterance('utt-3', 'file-3', 'speaker-2', 0, 1.5),
        rt.ExpUtterance('utt-4', 'file-3', 'speaker-2', 1.5, 2.5),
        rt.ExpUtterance('utt-5', 'file-4', 'speaker-3', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        'utt-1': [
            rt.ExpLabelList('text', 3),
            rt.ExpLabelList('raw_text', 1),
        ],
        'utt-2': [
            rt.ExpLabelList('text', 3),
            rt.ExpLabelList('raw_text', 1),
        ],
        'utt-3': [
            rt.ExpLabelList('text', 3),
            rt.ExpLabelList('raw_text', 1),
        ],
        'utt-4': [
            rt.ExpLabelList('text', 3),
            rt.ExpLabelList('raw_text', 1),
        ],
        'utt-5': [
            rt.ExpLabelList('text', 3),
            rt.ExpLabelList('raw_text', 1),
        ],
    }

    EXPECTED_LABELS = {
        'utt-1': [
            rt.ExpLabel('raw_text', 'who am i?', 0, float('inf')),
        ],
        'utt-4': [
            rt.ExpLabel('text', 'who', 0, float('inf')),
            rt.ExpLabel('text', 'are', 0, float('inf')),
            rt.ExpLabel('text', 'they', 3.5, 4.2),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 2
    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('train', [
            'utt-1',
            'utt-2',
            'utt-3',
        ]),
        rt.ExpSubview('dev', [
            'utt-4',
            'utt-5',
        ]),
    ]

    def test_load_issuers(self):
        ds = self.load()

        assert len(ds.issuers['speaker-3'].info) == 1
        assert ds.issuers['speaker-3'].info['region'] == 'zh'

    def test_load_label_meta(self):
        ds = self.load()

        utt_2 = ds.utterances['utt-2']
        utt_3 = ds.utterances['utt-3']
        utt_4 = ds.utterances['utt-4']

        utt_2_labels = sorted(utt_2.label_lists['text'].labels)
        utt_3_labels = sorted(utt_3.label_lists['text'].labels)
        utt_4_labels = sorted(utt_4.label_lists['text'].labels)

        assert len(utt_2_labels[1].meta) == 3
        assert utt_2_labels[1].meta['pron'] == 'huu'
        assert utt_2_labels[1].meta['duration'] == 2.3
        assert utt_2_labels[1].meta['stressed']

        assert len(utt_3_labels[0].meta) == 0

        assert len(utt_4_labels[2].meta) == 1
        assert utt_4_labels[2].meta['ex.'] == 19

    def test_load_features(self):
        ds = self.load()

        assert ds.feature_containers['mfcc'].path == os.path.join(self.SAMPLE_PATH, 'features', 'mfcc')
        assert ds.feature_containers['fbank'].path == os.path.join(self.SAMPLE_PATH, 'features', 'fbank')

    def load(self):
        return io.DefaultReader().load(self.SAMPLE_PATH)


class TestDefaultWriter:

    def test_save_files_exist(self, writer, sample_corpus, tmpdir):
        writer.save(sample_corpus, tmpdir.strpath)
        files = os.listdir(tmpdir.strpath)

        assert len(files) == 9

        assert 'files.txt' in files
        assert 'audio.txt' in files
        assert 'issuers.json' in files
        assert 'utterances.txt' in files
        assert 'utt_issuers.txt' in files
        assert 'labels_word-transcript.txt' in files
        assert 'subview_train.txt' in files
        assert 'subview_dev.txt' in files

    def test_save_file_tracks(self, writer, sample_corpus, tmpdir):
        # make sure relative path changes in contrast to self.ds.path
        out_path = os.path.join(tmpdir.strpath, 'somesubdir')
        os.makedirs(out_path)

        writer.save(sample_corpus, out_path)

        file_1_path = os.path.relpath(resources.sample_wav_file('wav_1.wav'), out_path)
        file_2_path = os.path.relpath(resources.sample_wav_file('wav_2.wav'), out_path)
        file_3_path = os.path.relpath(resources.sample_wav_file('wav_3.wav'), out_path)
        file_4_path = os.path.relpath(resources.sample_wav_file('wav_4.wav'), out_path)

        with open(os.path.join(out_path, 'files.txt'), 'r') as f:
            file_content = f.read()

        assert file_content.strip() == 'wav-1 {}\nwav_2 {}\nwav_3 {}\nwav_4 {}'.format(file_1_path,
                                                                                       file_2_path,
                                                                                       file_3_path,
                                                                                       file_4_path)

    def test_save_container_tracks(self, writer, tmpdir):
        # make sure relative path changes in contrast to self.ds.path
        out_path = os.path.join(tmpdir.strpath, 'somesubdir')
        os.makedirs(out_path)

        sample_corpus = io.DefaultReader().load(TestDefaultReader.SAMPLE_PATH)

        writer.save(sample_corpus, out_path)

        with open(os.path.join(out_path, 'audio.txt'), 'r') as f:
            file_content = f.read()

        rel_path = os.path.relpath(os.path.join(TestDefaultReader.SAMPLE_PATH, 'audio'), out_path)
        assert file_content.strip() == '\n'.join((
            'file-5 {} file-5'.format(rel_path),
            'file-6 {} file-6'.format(rel_path)
        ))

    def test_save_issuers(self, writer, sample_corpus, tmpdir):
        writer.save(sample_corpus, tmpdir.strpath)
        data = jsonfile.read_json_file(os.path.join(tmpdir.strpath, 'issuers.json'))

        expected = {
            'spk-1': {
                'type': 'speaker',
                'gender': 'male'
            },
            'spk-2': {
                'type': 'speaker',
                'gender': 'female'
            },
            'spk-3': {}
        }

        assert data == expected

    def test_save_utterances(self, writer, sample_corpus, tmpdir):
        writer.save(sample_corpus, tmpdir.strpath)

        with open(os.path.join(tmpdir.strpath, 'utterances.txt'), 'r') as f:
            file_content = f.read()

        assert file_content.strip() == 'utt-1 wav-1 0 -1\n' \
                                       'utt-2 wav_2 0 -1\n' \
                                       'utt-3 wav_3 0 1.5\n' \
                                       'utt-4 wav_3 1.5 2.5\n' \
                                       'utt-5 wav_4 0 -1'

    def test_save_utt_to_issuer(self, writer, sample_corpus, tmpdir):
        writer.save(sample_corpus, tmpdir.strpath)

        with open(os.path.join(tmpdir.strpath, 'utt_issuers.txt'), 'r') as f:
            file_content = f.read()

        assert file_content.strip() == 'utt-1 spk-1\n' \
                                       'utt-2 spk-1\n' \
                                       'utt-3 spk-2\n' \
                                       'utt-4 spk-2\n' \
                                       'utt-5 spk-3'

    def test_save_labels(self, writer, sample_corpus, tmpdir):
        writer.save(sample_corpus, tmpdir.strpath)

        with open(os.path.join(tmpdir.strpath, 'labels_word-transcript.txt'), 'r') as f:
            file_content = f.read()

        assert file_content.strip() == 'utt-1 0 -1 who am i\n' \
                                       'utt-2 0 -1 who are you [{"a": "hey", "b": 2}]\n' \
                                       'utt-3 0 -1 who is he\n' \
                                       'utt-4 0 -1 who are they\n' \
                                       'utt-5 0 -1 who is she'

    def test_save_subviews(self, writer, sample_corpus, tmpdir):
        writer.save(sample_corpus, tmpdir.strpath)

        with open(os.path.join(tmpdir.strpath, 'subview_train.txt'), 'r') as f:
            sv_train_content = f.read()

        with open(os.path.join(tmpdir.strpath, 'subview_dev.txt'), 'r') as f:
            sv_dev_content = f.read()

        assert sv_train_content.strip() == 'matching_utterance_ids\ninclude,utt-1,utt-2,utt-3'
        assert sv_dev_content.strip() == 'matching_utterance_ids\ninclude,utt-4,utt-5'

    def test_save_utterances_with_no_issuer(self, writer, sample_corpus, tmpdir):
        sample_corpus.utterances['utt-3'].issuer = None
        sample_corpus.utterances['utt-4'].issuer = None
        sample_corpus.utterances['utt-5'].issuer = None

        writer.save(sample_corpus, tmpdir.strpath)

        with open(os.path.join(tmpdir.strpath, 'utterances.txt'), 'r') as f:
            file_content = f.read()

        assert file_content.strip() == 'utt-1 wav-1 0 -1\n' \
                                       'utt-2 wav_2 0 -1\n' \
                                       'utt-3 wav_3 0 1.5\n' \
                                       'utt-4 wav_3 1.5 2.5\n' \
                                       'utt-5 wav_4 0 -1'

        with open(os.path.join(tmpdir.strpath, 'utt_issuers.txt'), 'r') as f:
            file_content = f.read()

        assert file_content.strip() == 'utt-1 spk-1\n' \
                                       'utt-2 spk-1'



"""
This module contains functions for handling different units.
Especially it provides function to convert from one to another unit (e.g. seconds -> sample-indexn).
"""

import math
import re

import numpy as np


def parse_storage_size(storage_size):
    """
    Parses an expression that represents an amount of storage/memory and returns the number of bytes it represents.

    Args:
        storage_size(str): Size in bytes. The units ``k`` (kibibytes), ``m`` (mebibytes) and ``g``
                           (gibibytes) are supported, i.e. a ``partition_size`` of ``1g`` equates :math:`2^{30}` bytes.

    Returns:
        int: Number of bytes.
    """
    pattern = re.compile(r'^([0-9]+(\.[0-9]+)?)([gmk])?$', re.I)

    units = {
        'k': 1024,
        'm': 1024 * 1024,
        'g': 1024 * 1024 * 1024
    }

    match = pattern.fullmatch(str(storage_size))

    if match is None:
        raise ValueError('Invalid partition size: {0}'.format(storage_size))

    groups = match.groups()

    # no units
    if groups[2] is None:
        # silently dropping the float, because byte is the smallest unit)
        return int(float(groups[0]))

    return int(float(groups[0]) * units[groups[2].lower()])


def seconds_to_sample(seconds, sampling_rate=16000):
    """
    Convert a value in seconds to a sample index based on the given sampling rate.

    Args:
        seconds (float): The value in seconds.
        sampling_rate (int): The sampling rate to use for conversion.

    Returns:
        int: The sample index (0-based).

    Example::
        >>> seconds_to_sample(1.3, sampling_rate=16000)
        20800
    """

    return int(np.round(sampling_rate * seconds))


def sample_to_seconds(sample, sampling_rate=16000):
    """
    Convert a sample index to seconds based on the given sampling rate.

    Args:
        sample (int): The index of the sample (0 based).
        sampling_rate (int): The sampling rate to use for conversion.

    Returns:
        float: The time in seconds.

    Example::
        >>> sample_to_seconds(20800, sampling_rate=16000)
        1.3
    """

    return sample / sampling_rate


class FrameSettings(object):
    """
    This class provides functions for handling conversions/calculations between time, samples and frames.

    By default the framing is done as follows:
        * The first frame starts at sample 0
        * The end of the last frame is higher than the last sample.
        * The end of the last frame is smaller than the last sample + hop_size

    Args:
        frame_size (int): Number of samples used per frame.
        hop_size (int): Number of samples between two frames.
    """

    def __init__(self, frame_size, hop_size):
        self.frame_size = frame_size
        self.hop_size = hop_size

    def num_frames(self, num_samples):
        """
        Return the number of frames that will be used for a signal with the length of ``num_samples``.
        """
        return math.ceil(float(max(num_samples - self.frame_size, 0)) / float(self.hop_size)) + 1

    def sample_to_frame_range(self, sample_index):
        """
        Return a tuple containing the indices of the first frame containing the sample with the given index and
        the last frame (exclusive, doesn't contain the sample anymore).
        """
        start = max(0, int((sample_index - self.frame_size) / self.hop_size) + 1)
        end = int(sample_index / self.hop_size) + 1
        return start, end

    def frame_to_sample(self, frame_index):
        """
        Return a tuple containing the indices of the sample which are the first sample and the end (exclusive)
        of the frame with the given index.
        """
        start = frame_index * self.hop_size
        end = start + self.frame_size
        return start, end

    def frame_to_seconds(self, frame_index, sr):
        """
        Return a tuple containing the start and end of the frame in seconds.
        """
        start_sample, end_sample = self.frame_to_sample(frame_index)
        return sample_to_seconds(start_sample, sampling_rate=sr), sample_to_seconds(end_sample, sampling_rate=sr)

    def time_range_to_frame_range(self, start, end, sr):
        """
        Calculate the frames containing samples from the given time range in seconds.

        Args:
            start (float): Start time in seconds.
            end (float): End time in seconds.
            sr (int): The sampling rate to use for time-to-sample conversion.

        Returns:
            tuple: A tuple containing the start and end (exclusive) frame indices.
        """

        start_sample = seconds_to_sample(start, sr)
        end_sample = seconds_to_sample(end, sr)

        return self.sample_to_frame_range(start_sample)[0], self.sample_to_frame_range(end_sample - 1)[1]



import os

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus import io

from tests import resources
from . import reader_test as rt


class TestTudaReader(rt.CorpusReaderTest):

    # SHORT aliases for speakers
    SPK_cf = 'cf372280-5606-4b05-9d24-3ab7805d8462'
    SPK_55 = '55065c47-1290-4974-997e-e77f24e7c72d'
    SPK_75 = '755d9b71-f36e-45a6-a437-edebcfaee08d'
    SPK_9e = '9e6a00c9-80f0-479d-8b36-4139a9571217'
    SPK_58 = '58b8b441-684f-4753-aa16-589f1e149fa0'
    SPK_2a = '2a0995a7-47d8-453f-9864-5940efd3c71a'
    SPK_40 = '40a95aaf-2d87-43dc-b00e-9a4ceb77c6db'

    SAMPLE_PATH = resources.sample_corpus_path('tuda')
    FILE_TRACK_BASE_PATH = SAMPLE_PATH

    EXPECTED_NUMBER_OF_TRACKS = 48
    EXPECTED_TRACKS = [
        # DEV
        rt.ExpFileTrack('2015-01-27-11-31-32_Kinect-Beam', 'dev/2015-01-27-11-31-32_Kinect-Beam.wav'),
        rt.ExpFileTrack('2015-01-27-11-31-32_Kinect-RAW', 'dev/2015-01-27-11-31-32_Kinect-RAW.wav'),
        rt.ExpFileTrack('2015-01-27-11-31-32_Realtek', 'dev/2015-01-27-11-31-32_Realtek.wav'),
        rt.ExpFileTrack('2015-01-27-11-31-32_Samson', 'dev/2015-01-27-11-31-32_Samson.wav'),
        rt.ExpFileTrack('2015-01-28-12-36-24_Yamaha', 'dev/2015-01-28-12-36-24_Yamaha.wav'),
        # TEST
        rt.ExpFileTrack('2015-01-27-12-34-36_Kinect-Beam', 'test/2015-01-27-12-34-36_Kinect-Beam.wav'),
        rt.ExpFileTrack('2015-01-27-12-34-36_Kinect-RAW', 'test/2015-01-27-12-34-36_Kinect-RAW.wav'),
        rt.ExpFileTrack('2015-01-27-12-34-36_Realtek', 'test/2015-01-27-12-34-36_Realtek.wav'),
        rt.ExpFileTrack('2015-01-27-12-34-36_Samson', 'test/2015-01-27-12-34-36_Samson.wav'),
        rt.ExpFileTrack('2015-01-27-12-34-36_Yamaha', 'test/2015-01-27-12-34-36_Yamaha.wav'),
        # TRAIN
        rt.ExpFileTrack('2014-03-17-10-26-07_Microsoft-Kinect-Raw',
                        'train/2014-03-17-10-26-07_Microsoft-Kinect-Raw.wav'),
        rt.ExpFileTrack('2014-03-17-10-26-07_Realtek', 'train/2014-03-17-10-26-07_Realtek.wav'),
        rt.ExpFileTrack('2014-03-17-10-26-07_Yamaha', 'train/2014-03-17-10-26-07_Yamaha.wav'),
        rt.ExpFileTrack('2014-03-17-13-03-33_Kinect-Beam', 'train/2014-03-17-13-03-33_Kinect-Beam.wav'),
        rt.ExpFileTrack('2014-03-17-13-03-33_Realtek', 'train/2014-03-17-13-03-33_Realtek.wav'),
        rt.ExpFileTrack('2014-03-17-13-03-33_Yamaha', 'train/2014-03-17-13-03-33_Yamaha.wav'),
        rt.ExpFileTrack('2014-03-19-15-01-56_Kinect-Beam', 'train/2014-03-19-15-01-56_Kinect-Beam.wav'),
        # NOT ALL LISTED
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 7
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker(SPK_cf, 8, issuers.Gender.MALE, issuers.AgeGroup.ADULT, 'deu'),
        rt.ExpSpeaker(SPK_55, 10, issuers.Gender.MALE, issuers.AgeGroup.ADULT, 'deu'),
        rt.ExpSpeaker(SPK_75, 7, issuers.Gender.MALE, issuers.AgeGroup.ADULT, 'deu'),
        rt.ExpSpeaker(SPK_9e, 5, issuers.Gender.FEMALE, issuers.AgeGroup.YOUTH, 'deu'),
        rt.ExpSpeaker(SPK_58, 10, issuers.Gender.FEMALE, issuers.AgeGroup.ADULT, 'deu'),
        rt.ExpSpeaker(SPK_2a, 5, issuers.Gender.MALE, issuers.AgeGroup.ADULT, 'deu'),
        rt.ExpSpeaker(SPK_40, 3, issuers.Gender.MALE, issuers.AgeGroup.ADULT, 'deu'),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 48
    EXPECTED_UTTERANCES = [
        # DEV
        rt.ExpUtterance('2015-01-27-11-31-32_Kinect-Beam', '2015-01-27-11-31-32_Kinect-Beam', SPK_75, 0, float('inf')),
        rt.ExpUtterance('2015-01-27-11-31-32_Kinect-RAW', '2015-01-27-11-31-32_Kinect-RAW', SPK_75, 0, float('inf')),
        rt.ExpUtterance('2015-01-27-11-31-32_Realtek', '2015-01-27-11-31-32_Realtek', SPK_75, 0, float('inf')),
        rt.ExpUtterance('2015-01-27-11-31-32_Samson', '2015-01-27-11-31-32_Samson', SPK_75, 0, float('inf')),
        rt.ExpUtterance('2015-01-28-12-36-24_Yamaha', '2015-01-28-12-36-24_Yamaha', SPK_9e, 0, float('inf')),
        # TEST
        rt.ExpUtterance('2015-01-27-12-34-36_Kinect-Beam', '2015-01-27-12-34-36_Kinect-Beam', SPK_58, 0, float('inf')),
        rt.ExpUtterance('2015-01-27-12-34-36_Kinect-RAW', '2015-01-27-12-34-36_Kinect-RAW', SPK_58, 0, float('inf')),
        rt.ExpUtterance('2015-01-27-12-34-36_Realtek', '2015-01-27-12-34-36_Realtek', SPK_58, 0, float('inf')),
        rt.ExpUtterance('2015-01-27-12-34-36_Samson', '2015-01-27-12-34-36_Samson', SPK_58, 0, float('inf')),
        rt.ExpUtterance('2015-01-27-12-34-36_Yamaha', '2015-01-27-12-34-36_Yamaha', SPK_58, 0, float('inf')),
        # TRAIN
        rt.ExpUtterance('2014-03-17-10-26-07_Microsoft-Kinect-Raw',
                        '2014-03-17-10-26-07_Microsoft-Kinect-Raw', SPK_40, 0, float('inf')),
        rt.ExpUtterance('2014-03-17-10-26-07_Realtek', '2014-03-17-10-26-07_Realtek', SPK_40, 0, float('inf')),
        rt.ExpUtterance('2014-03-17-10-26-07_Yamaha', '2014-03-17-10-26-07_Yamaha', SPK_40, 0, float('inf')),
        rt.ExpUtterance('2014-03-17-13-03-33_Kinect-Beam', '2014-03-17-13-03-33_Kinect-Beam', SPK_cf, 0, float('inf')),
        rt.ExpUtterance('2014-03-17-13-03-33_Realtek', '2014-03-17-13-03-33_Realtek', SPK_cf, 0, float('inf')),
        rt.ExpUtterance('2014-03-17-13-03-33_Yamaha', '2014-03-17-13-03-33_Yamaha', SPK_cf, 0, float('inf')),
        rt.ExpUtterance('2014-03-19-15-01-56_Kinect-Beam', '2014-03-19-15-01-56_Kinect-Beam', SPK_cf, 0, float('inf')),
        # NOT ALL LISTED
    ]

    EXPECTED_LABEL_LISTS = {
        '2015-01-27-11-31-32_Samson': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1),
        ],
        '2015-01-27-11-31-32_Realtek': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1),
        ],
        '2014-03-17-10-26-07_Realtek': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT_RAW, 1),
        ],
        # NOT ALL LISTED
    }

    EXPECTED_LABELS = {
        '2015-01-27-11-31-32_Samson': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Manche haben dass', 0, float('inf')),
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT_RAW, 'Manche haben , dass.', 0, float('inf')),
        ],
        '2015-01-27-11-31-32_Realtek': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Manche haben dass', 0, float('inf')),
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT_RAW, 'Manche haben , dass.', 0, float('inf')),
        ],
        '2014-03-17-13-03-33_Realtek': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'Ich habe mich', 0, float('inf')),
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT_RAW, 'Ich habe mich.', 0, float('inf')),
        ],
        # NOT ALL LISTED
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 3 + 6 + 5 + 5 + 6
    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('test', [
            '2015-01-27-12-34-36_Kinect-Beam',
            '2015-01-27-12-34-36_Kinect-RAW',
            '2015-01-27-12-34-36_Realtek',
            '2015-01-27-12-34-36_Samson',
            '2015-01-27-12-34-36_Yamaha',
            '2015-02-03-12-08-13_Kinect-Beam',
            '2015-02-03-12-08-13_Kinect-RAW',
            '2015-02-03-12-08-13_Realtek',
            '2015-02-03-12-08-13_Samson',
            '2015-02-03-12-08-13_Yamaha',
            '2015-02-10-14-31-52_Kinect-Beam',
            '2015-02-10-14-31-52_Kinect-RAW',
            '2015-02-10-14-31-52_Realtek',
            '2015-02-10-14-31-52_Samson',
            '2015-02-10-14-31-52_Yamaha',
        ]),
        rt.ExpSubview('Microsoft-Kinect-Raw', [
            '2014-03-17-10-26-07_Microsoft-Kinect-Raw',
            '2014-03-17-13-03-33_Microsoft-Kinect-Raw',
        ]),
        rt.ExpSubview('train_Samson', [
            '2014-08-07-13-22-38_Samson',
            '2014-08-14-14-52-00_Samson',
        ]),
        # NOT ALL LISTED
    ]

    def load(self):
        reader = io.TudaReader()
        return reader.load(self.SAMPLE_PATH)

    def test_get_ids_from_folder(self):
        assert io.TudaReader.get_ids_from_folder(os.path.join(self.SAMPLE_PATH, 'train'), 'train') == {
            '2014-03-17-10-26-07',
            '2014-03-17-13-03-33',
            '2014-03-19-15-01-56',
            '2014-08-07-13-22-38',
            '2014-08-14-14-52-00',
            '2015-04-17-11-26-07',
        }
        assert io.TudaReader.get_ids_from_folder(os.path.join(self.SAMPLE_PATH, 'dev'), 'dev') == {
            '2015-01-27-11-31-32',
            '2015-01-28-11-35-47',
            '2015-01-28-12-36-24',
        }
        assert io.TudaReader.get_ids_from_folder(os.path.join(self.SAMPLE_PATH, 'test'), 'test') == {
            '2015-01-27-12-34-36',
            '2015-02-03-12-08-13',
            '2015-02-10-14-31-52',
        }

    def test_get_ids_from_folder_ignore_bad_files(self):
        ids = io.TudaReader.get_ids_from_folder(os.path.join(self.SAMPLE_PATH, 'train'), 'train')
        assert '2014-08-05-11-08-34' not in ids

        ids = io.TudaReader.get_ids_from_folder(os.path.join(self.SAMPLE_PATH, 'dev'), 'dev')
        assert '2015-01-28-11-49-53' not in ids



import numpy as np

from . import base


class AvgPool(base.Computation):
    """
    Average a given number of sequential frames into a single frame.
    If at the end of a stream just the remaining frames are used, no matter how many there are left.

    Args:
        size (float): The maximum number of frames to pool by taking the mean.
    """

    def __init__(self, size, parent=None, name=None):
        super(AvgPool, self).__init__(min_frames=size, parent=parent, name=name)

        self.size = size
        self.rest = None

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        # Get rid of old rest frames
        if chunk.offset == 0:
            self.rest = None

        # Merge incoming plus rest from previous chunk
        all = chunk.data

        if self.rest is not None:
            all = np.vstack([self.rest, all])

        # Only use a multiple of self.size
        num_rest = all.shape[0] % self.size

        if num_rest > 0:
            self.rest = all[-num_rest:, :]
            all = all[:-num_rest, :]
        else:
            self.rest = None

        # Average
        all_mean = np.mean(all.reshape(-1, self.size, all.shape[1]), axis=1)

        if chunk.is_last and self.rest is not None:
            rest_mean = np.mean(self.rest, axis=0)
            all_mean = np.vstack([all_mean, rest_mean])

        return all_mean

    def frame_transform_step(self, frame_size, hop_size):
        fs_transformed = frame_size + (self.size - 1) * hop_size
        hs_transformed = self.size * hop_size

        return fs_transformed, hs_transformed


class VarPool(base.Computation):
    """
    Variance over a given number of sequential frames to form a single frame.
    If at the end of a stream just the remaining frames are used, no matter how many there are left.

    Args:
        size (float): The maximum number of frames to pool by taking the mean.
    """

    def __init__(self, size, parent=None, name=None):
        super(VarPool, self).__init__(min_frames=size, parent=parent, name=name)

        self.size = size
        self.rest = None

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        # Get rid of old rest frames
        if chunk.offset == 0:
            self.rest = None

        # Merge incoming plus rest from previous chunk
        all = chunk.data

        if self.rest is not None:
            all = np.vstack([self.rest, all])

        # Only use a multiple of self.size
        num_rest = all.shape[0] % self.size

        if num_rest > 0:
            self.rest = all[-num_rest:, :]
            all = all[:-num_rest, :]
        else:
            self.rest = None

        # Average
        all_var = np.var(all.reshape(-1, self.size, all.shape[1]), axis=1)

        if chunk.is_last and self.rest is not None:
            rest_var = np.var(self.rest, axis=0)
            all_var = np.vstack([all_var, rest_var])

        return all_var

    def frame_transform_step(self, frame_size, hop_size):
        fs_transformed = frame_size + (self.size - 1) * hop_size
        hs_transformed = self.size * hop_size

        return fs_transformed, hs_transformed



import os

import pytest
import requests_mock

from audiomate import corpus
from audiomate.corpus import io
from audiomate.corpus.io import esc
from tests import resources

from . import reader_test as rt


@pytest.fixture
def downloader():
    return io.ESC50Downloader()


@pytest.fixture()
def zip_data():
    path = resources.get_resource_path(
        ['sample_files', 'zip_sample_with_subfolder.zip']
    )
    with open(path, 'rb') as f:
        return f.read()


class TestESC50Downloader:

    def test_download(self, zip_data, downloader, tmpdir):
        target_folder = tmpdir.strpath

        with requests_mock.Mocker() as mock:
            mock.get(esc.DOWNLOAD_URL, content=zip_data)

            downloader.download(target_folder)

        assert os.path.isfile(os.path.join(target_folder, 'a.txt'))
        assert os.path.isfile(os.path.join(target_folder, 'subsub', 'b.txt'))
        assert os.path.isfile(os.path.join(target_folder, 'subsub', 'c.txt'))


class TestESC50Reader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('esc50')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'audio')

    EXPECTED_NUMBER_OF_TRACKS = 10
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('1-119125-A-45', '1-119125-A-45.wav'),
        rt.ExpFileTrack('1-12654-B-15', '1-12654-B-15.wav'),
        rt.ExpFileTrack('1-155858-E-25',  '1-155858-E-25.wav'),
        rt.ExpFileTrack('1-155858-F-25',  '1-155858-F-25.wav'),
        rt.ExpFileTrack('1-15689-A-4',  '1-15689-A-4.wav'),
        rt.ExpFileTrack('1-17124-A-43',  '1-17124-A-43.wav'),
        rt.ExpFileTrack('1-18755-A-4',  '1-18755-A-4.wav'),
        rt.ExpFileTrack('1-18755-B-4',  '1-18755-B-4.wav'),
        rt.ExpFileTrack('1-18757-A-4',  '1-18757-A-4.wav'),
        rt.ExpFileTrack('1-18810-A-49',  '1-18810-A-49.wav')
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 10
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('1-119125-A-45', '1-119125-A-45', None, 0, float('inf')),
        rt.ExpUtterance('1-12654-B-15', '1-12654-B-15', None, 0, float('inf')),
        rt.ExpUtterance('1-155858-E-25', '1-155858-E-25', None, 0, float('inf')),
        rt.ExpUtterance('1-155858-F-25', '1-155858-F-25', None, 0, float('inf')),
        rt.ExpUtterance('1-15689-A-4', '1-15689-A-4', None, 0, float('inf')),
        rt.ExpUtterance('1-17124-A-43', '1-17124-A-43', None, 0, float('inf')),
        rt.ExpUtterance('1-18755-A-4', '1-18755-A-4', None, 0, float('inf')),
        rt.ExpUtterance('1-18755-B-4', '1-18755-B-4', None, 0, float('inf')),
        rt.ExpUtterance('1-18757-A-4', '1-18757-A-4', None, 0, float('inf')),
        rt.ExpUtterance('1-18810-A-49', '1-18810-A-49', None, 0, float('inf')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 0

    EXPECTED_LABEL_LISTS = {
        '1-119125-A-45': [
            rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1),
        ],
        '1-15689-A-4': [
            rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1),
        ],
        '1-18810-A-49': [
            rt.ExpLabelList(corpus.LL_SOUND_CLASS, 1),
        ],
    }

    EXPECTED_LABELS = {
        '1-119125-A-45': [
            rt.ExpLabel(corpus.LL_SOUND_CLASS, 'train', 0, float('inf')),
        ],
        '1-15689-A-4': [
            rt.ExpLabel(corpus.LL_SOUND_CLASS, 'frog', 0, float('inf')),
        ],
        '1-18810-A-49': [
            rt.ExpLabel(corpus.LL_SOUND_CLASS, 'hand_saw', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 6
    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('fold-1', ['1-119125-A-45', '1-12654-B-15',
                                 '1-155858-F-25', '1-17124-A-43',
                                 '1-18755-A-4']),
        rt.ExpSubview('fold-2', ['1-155858-E-25', '1-15689-A-4']),
        rt.ExpSubview('fold-3', ['1-18755-B-4']),
        rt.ExpSubview('fold-4', ['1-18757-A-4']),
        rt.ExpSubview('fold-5', ['1-18810-A-49']),
        rt.ExpSubview('esc-10', ['1-155858-E-25', '1-155858-F-25',
                                 '1-17124-A-43']),
    ]

    def load(self):
        return io.ESC50Reader().load(self.SAMPLE_PATH)



import numpy as np

from . import base
from audiomate.utils import misc


class FrameHotEncoder(base.Encoder):
    """
    The FrameHotEncoder is used to encode the labels per frame.
    It creates a matrix with dimension num-frames x len(labels).
    The vector (2nd dim) has an entry for every label in the passed labels-list.
    If the sequence contains a given label within a frame it is set to 1.

    Arguments:
        labels (list): List of labels (str) which should be included in the vector representation.
        label_list_idx (str): The name of the label-list to use for encoding.
                              Only labels of this label-list are considered.
        frame_settings (FrameSettings): Frame settings to use.
        sr (int): The sampling rate used, if None it is assumed the native sampling rate from the file is used.

    Example:
        >>> from audiomate import annotations
        >>> from audiomate.utils import units import
        >>>
        >>> ll = annotations.LabelList(idx='test', labels=[
        >>>     annotations.Label('music', 0, 2),
        >>>     annotations.Label('speech', 2, 5),
        >>>     annotations.Label('noise', 4, 6),
        >>>     annotations.Label('music', 6, 8)
        >>> ])
        >>> utt.set_label_list(ll)
        >>>
        >>> labels = ['speech', 'music', 'noise']
        >>> fs = units.FrameSettings(16000, 16000)
        >>> encoder = FrameHotEncoder(labels, 'test', frame_settings=fs, sr=16000)
        >>> encoder.encode_utterance(utt)
        array([
            [0, 1, 0],
            [0, 1, 0],
            [1, 0, 0],
            [1, 0, 0],
            [1, 0, 1],
            [0, 0, 1],
            [0, 1, 0],
            [0, 1, 0]
        ])

    """

    def __init__(self, labels, label_list_idx, frame_settings, sr=None):
        self.labels = labels
        self.label_list_idx = label_list_idx
        self.frame_settings = frame_settings
        self.sr = sr

    def encode_utterance(self, utterance, corpus=None):
        sr = self.sr or utterance.sampling_rate
        num_samples = utterance.num_samples(sr=sr)
        num_frames = self.frame_settings.num_frames(num_samples)

        mat = np.zeros((num_frames, len(self.labels)))

        if self.label_list_idx not in utterance.label_lists:
            raise ValueError('Utterance {} has no label-list with idx {}'.format(utterance.idx, self.label_list_idx))

        label_list = utterance.label_lists[self.label_list_idx]

        for label in label_list:
            if label.value in self.labels:
                if label.end == float('inf'):
                    if utterance.end == float('inf'):
                        label_end = utterance.track.duration
                    else:
                        label_end = utterance.duration
                else:
                    label_end = label.end

                start, end = self.frame_settings.time_range_to_frame_range(label.start, label_end, sr)

                # If label ends at the end of the utterance
                if label.end == float('inf'):
                    end = num_frames

                mat[start:end, self.labels.index(label.value)] = 1

        return mat


class FrameOrdinalEncoder(base.Encoder):
    """
    The FrameOrdinalEncoder is used to encode the labels per frame.
    It creates a vector with length num-frames.
    For every frame sets the index of the label that is present for that frame.
    If multiple labels are present the longest within the frame.
    If multiple labels have the same length the smaller index is selected, hence
    the passed `labels` list acts as a priority.

    Arguments:
        labels (list): List of labels (str) which should be included in the vector representation.
        label_list_idx (str): The name of the label-list to use for encoding.
                              Only labels of this label-list are considered.
        frame_settings (FrameSettings): Frame settings to use.
        sr (int): The sampling rate used, if None it is assumed the native sampling rate from the file is used.

    Example:
        >>> from audiomate import annotations
        >>> from audiomate.utils import units import
        >>>
        >>> ll = annotations.LabelList(idx='test', labels=[
        >>>     annotations.Label('music', 0, 2),
        >>>     annotations.Label('speech', 2, 5),
        >>>     annotations.Label('noise', 4, 6),
        >>>     annotations.Label('music', 6, 8)
        >>> ])
        >>> utt.set_label_list(ll)
        >>>
        >>> labels = ['speech', 'music', 'noise']
        >>> fs = units.FrameSettings(16000, 16000)
        >>> encoder = FrameOrdinalEncoder(labels, 'test', frame_settings=fs)
        >>> encoder.encode_utterance(utt)
        array([1,1,0,0,0,2,1,1])
    """

    def __init__(self, labels, label_list_idx, frame_settings, sr=None):
        self.labels = labels
        self.label_list_idx = label_list_idx
        self.frame_settings = frame_settings
        self.sr = sr

    def encode_utterance(self, utterance, corpus=None):
        sr = self.sr or utterance.sampling_rate
        num_samples = utterance.num_samples(sr=sr)
        num_frames = self.frame_settings.num_frames(num_samples)

        mat = np.zeros((num_frames, len(self.labels)))

        if self.label_list_idx not in utterance.label_lists:
            raise ValueError('Utterance {} has no label-list with idx {}'.format(utterance.idx, self.label_list_idx))

        label_list = utterance.label_lists[self.label_list_idx]

        for label in label_list:
            if label.value in self.labels:
                if label.end == float('inf'):
                    if utterance.end == float('inf'):
                        label_end = utterance.track.duration
                    else:
                        label_end = utterance.duration
                else:
                    label_end = label.end

                start, end = self.frame_settings.time_range_to_frame_range(label.start, label_end, sr)

                # If label ends at the end of the utterance
                if label.end == float('inf'):
                    end = num_frames

                label_index = self.labels.index(label.value)

                for frame_index in range(start, min(end, num_frames)):
                    frame_start, frame_end = self.frame_settings.frame_to_seconds(frame_index, sr=sr)
                    overlap = misc.length_of_overlap(frame_start, frame_end, label.start, label.end)

                    mat[frame_index, label_index] = overlap

        return np.argmax(mat, axis=1)



import copy


class Issuer(object):
    """
    The issuer represents a person, object or something that produced an
    utterance. Technically the issuer can be used to group utterances
    that came from the same source.

    Args:
        idx (str): An unique identifier for this issuer within a dataset.
        info (dict): Any additional info for this issuer as dict.

    Attributes:
        Issuer.utterances (list): List of utterances that this issuer owns.
    """

    __slots__ = ['idx', 'info', 'utterances']

    def __init__(self, idx, info={}):
        self.idx = idx
        self.info = info
        self.utterances = set()

    def __str__(self):
        return 'Issuer(idx={0}, info={1})'.format(self.idx, self.info)

    def __copy__(self):
        # self.utterances is ignored intentionally
        # only a "weak-ref" when added to a corpus

        cp = Issuer(
            self.idx,
            info=self.info
        )

        return cp

    def __deepcopy__(self, memo):
        # self.utterances is ignored intentionally
        # only a "weak-ref" when added to a corpus

        cp = Issuer(
            self.idx,
            info=copy.deepcopy(self.info, memo)
        )

        return cp



import collections
import glob
import os
import re
import json

import audiomate
from audiomate import containers
from audiomate import tracks
from audiomate import annotations
from audiomate import issuers
from audiomate.corpus.subset import subview
from audiomate.utils import textfile
from audiomate.utils import jsonfile
from . import base

FILES_FILE_NAME = 'files.txt'
ISSUER_FILE_NAME = 'issuers.json'
UTTERANCE_FILE_NAME = 'utterances.txt'
UTT_ISSUER_FILE_NAME = 'utt_issuers.txt'
LABEL_FILE_PREFIX = 'labels'
FEAT_CONTAINER_FILE_NAME = 'features.txt'
AUDIO_CONTAINER_FILE_NAME = 'audio.txt'
SUBVIEW_FILE_PREFIX = 'subview'

LABEL_META_REGEX = r'(.*) \[(\{.*\})\]'
META_PATTERN = re.compile(LABEL_META_REGEX)


class DefaultReader(base.CorpusReader):
    """
    Reads corpora in the Default format.
    """

    @classmethod
    def type(cls):
        return 'default'

    def _check_for_missing_files(self, path):
        necessary_files = [FILES_FILE_NAME, UTTERANCE_FILE_NAME]
        missing_files = []

        for file_name in necessary_files:
            file_path = os.path.join(path, file_name)

            if not os.path.isfile(file_path):
                missing_files.append(file_name)

        return missing_files

    def _load(self, path):
        file_path = os.path.join(path, FILES_FILE_NAME)
        audio_path = os.path.join(path, AUDIO_CONTAINER_FILE_NAME)
        issuer_path = os.path.join(path, ISSUER_FILE_NAME)
        utt_issuer_path = os.path.join(path, UTT_ISSUER_FILE_NAME)
        utterance_path = os.path.join(path, UTTERANCE_FILE_NAME)
        feat_path = os.path.join(path, FEAT_CONTAINER_FILE_NAME)

        corpus = audiomate.Corpus(path=path)

        DefaultReader.read_files(file_path, corpus)
        DefaultReader.read_tracks_from_audio_containers(audio_path, corpus)
        DefaultReader.read_issuers(issuer_path, corpus)
        utt_id_to_issuer = DefaultReader.read_utt_to_issuer_mapping(utt_issuer_path, corpus)
        DefaultReader.read_utterances(utterance_path, corpus, utt_id_to_issuer)
        DefaultReader.read_labels(path, corpus)
        DefaultReader.read_feature_containers(feat_path, corpus)
        DefaultReader.read_subviews(path, corpus)

        return corpus

    @staticmethod
    def read_files(file_path, corpus):
        path = os.path.dirname(file_path)
        for file_idx, file_path in textfile.read_key_value_lines(file_path, separator=' ').items():
            corpus.new_file(os.path.join(path, file_path), track_idx=file_idx, copy_file=False)

    @staticmethod
    def read_issuers(file_path, corpus):
        if not os.path.isfile(file_path):
            return

        data = jsonfile.read_json_file(file_path)

        for issuer_idx, issuer_data in data.items():
            issuer_type = issuer_data.get('type', None)
            issuer_info = issuer_data.get('info', {})

            if issuer_type == 'speaker':
                gender = issuers.Gender(issuer_data.get('gender', 'unknown').lower())
                age_group = issuers.AgeGroup(issuer_data.get('age_group', 'unknown').lower())
                native_language = issuer_data.get('native_language', None)

                issuer = issuers.Speaker(issuer_idx,
                                         gender=gender,
                                         age_group=age_group,
                                         native_language=native_language,
                                         info=issuer_info)
            elif issuer_type == 'artist':
                name = issuer_data.get('name', None)

                issuer = issuers.Artist(issuer_idx,
                                        name=name,
                                        info=issuer_info)
            else:
                issuer = issuers.Issuer(issuer_idx, info=issuer_info)

            corpus.import_issuers(issuer)

    @staticmethod
    def read_utt_to_issuer_mapping(utt_issuer_path, corpus):
        utt_issuers = {}

        if os.path.isfile(utt_issuer_path):
            for utt_id, issuer_idx in textfile.read_key_value_lines(utt_issuer_path, separator=' ').items():
                if issuer_idx in corpus.issuers.keys():
                    utt_issuers[utt_id] = corpus.issuers[issuer_idx]
                else:
                    utt_issuers[utt_id] = corpus.new_issuer(issuer_idx=issuer_idx)

        return utt_issuers

    @staticmethod
    def read_utterances(utterance_path, corpus, utt_idx_to_issuer):
        utterances = textfile.read_separated_lines_with_first_key(utterance_path, separator=' ', max_columns=4)

        for utterance_idx, utt_info in utterances.items():
            issuer_idx = None
            start = 0
            end = float('inf')

            if len(utt_info) > 1:
                start = float(utt_info[1])

            if len(utt_info) > 2:
                end = float(utt_info[2])

                if end == -1:
                    end = float('inf')

            if utterance_idx in utt_idx_to_issuer.keys():
                issuer_idx = utt_idx_to_issuer[utterance_idx].idx

            corpus.new_utterance(utterance_idx, utt_info[0], issuer_idx=issuer_idx, start=start, end=end)

    @staticmethod
    def read_labels(path, corpus):

        for label_file in glob.glob(os.path.join(path, '{}_*.txt'.format(LABEL_FILE_PREFIX))):
            file_name = os.path.basename(label_file)
            key = file_name[len('{}_'.format(LABEL_FILE_PREFIX)):len(file_name) - len('.txt')]

            utterance_labels = collections.defaultdict(list)

            labels = textfile.read_separated_lines_generator(label_file, separator=' ', max_columns=4)

            for record in labels:
                label = record[3]
                start = float(record[1])
                end = float(record[2])
                meta = None
                meta_match = META_PATTERN.match(label)

                if end == -1:
                    end = float('inf')

                if meta_match is not None:
                    meta_json = meta_match.group(2)
                    meta = json.loads(meta_json)
                    label = meta_match.group(1)

                utterance_labels[record[0]].append(annotations.Label(label, start, end, meta=meta))

            for utterance_idx, labels in utterance_labels.items():
                ll = annotations.LabelList(idx=key, labels=labels)
                corpus.utterances[utterance_idx].set_label_list(ll)

    @staticmethod
    def read_feature_containers(feat_path, corpus):
        if os.path.isfile(feat_path):
            base_path = os.path.dirname(feat_path)
            containers = textfile.read_key_value_lines(feat_path, separator=' ')
            for container_name, container_path in containers.items():
                corpus.new_feature_container(container_name, path=os.path.join(base_path, container_path))

    @staticmethod
    def read_tracks_from_audio_containers(audio_path, corpus):
        if os.path.isfile(audio_path):
            base_path = os.path.dirname(audio_path)
            audio_tracks = textfile.read_separated_lines(audio_path,
                                                         separator=' ',
                                                         max_columns=3)

            audio_containers = {}

            for entry in audio_tracks:
                track_idx = entry[0]
                container_path = entry[1]
                key = entry[2]

                if container_path in audio_containers.keys():
                    container = audio_containers[key]
                else:
                    abs_path = os.path.abspath(os.path.join(base_path, container_path))
                    container = containers.AudioContainer(abs_path)

                track = tracks.ContainerTrack(track_idx, container, key)
                corpus.import_tracks(track)

    @staticmethod
    def read_subviews(path, corpus):
        for sv_file in glob.glob(os.path.join(path, '{}_*.txt'.format(SUBVIEW_FILE_PREFIX))):
            file_name = os.path.basename(sv_file)
            key = file_name[len('{}_'.format(SUBVIEW_FILE_PREFIX)):len(file_name) - len('.txt')]

            with open(sv_file, 'r') as f:
                content = f.read().strip()

            sv = subview.Subview.parse(content)
            corpus.import_subview(key, sv)


class DefaultWriter(base.CorpusWriter):
    """
    Writes corpora in the Default format.
    """

    @classmethod
    def type(cls):
        return 'default'

    def _save(self, corpus, path):
        file_path = os.path.join(path, FILES_FILE_NAME)
        audio_path = os.path.join(path, AUDIO_CONTAINER_FILE_NAME)
        issuer_path = os.path.join(path, ISSUER_FILE_NAME)
        utterance_path = os.path.join(path, UTTERANCE_FILE_NAME)
        utt_issuer_path = os.path.join(path, UTT_ISSUER_FILE_NAME)
        container_path = os.path.join(path, FEAT_CONTAINER_FILE_NAME)

        DefaultWriter.write_file_tracks(file_path, corpus, path)
        DefaultWriter.write_container_tracks(audio_path, corpus, path)
        DefaultWriter.write_issuers(issuer_path, corpus)
        DefaultWriter.write_utterances(utterance_path, corpus)
        DefaultWriter.write_utt_to_issuer_mapping(utt_issuer_path, corpus)
        DefaultWriter.write_labels(path, corpus)
        DefaultWriter.write_feature_containers(container_path, corpus)
        DefaultWriter.write_subviews(path, corpus)

    @staticmethod
    def write_file_tracks(file_path, corpus, path):
        file_records = []

        for file in corpus.tracks.values():
            if isinstance(file, tracks.FileTrack):
                file_records.append([
                    file.idx,
                    os.path.relpath(file.path, path)
                ])

        textfile.write_separated_lines(file_path, file_records, separator=' ', sort_by_column=0)

    @staticmethod
    def write_container_tracks(audio_path, corpus, path):
        container_records = set({})

        for track in corpus.tracks.values():
            if isinstance(track, tracks.ContainerTrack):
                rel_path = os.path.relpath(track.container.path, path)
                container_records.add((
                    track.idx,
                    rel_path,
                    track.key
                ))

        textfile.write_separated_lines(
            audio_path,
            container_records,
            separator=' ',
            sort_by_column=0
        )

    @staticmethod
    def write_issuers(file_path, corpus):
        data = {}

        for issuer in corpus.issuers.values():
            issuer_data = {}

            if issuer.info is not None and len(issuer.info) > 0:
                issuer_data['info'] = issuer.info

            if type(issuer) == issuers.Speaker:
                issuer_data['type'] = 'speaker'

                if issuer.gender != issuers.Gender.UNKNOWN:
                    issuer_data['gender'] = issuer.gender.value

                if issuer.age_group != issuers.AgeGroup.UNKNOWN:
                    issuer_data['age_group'] = issuer.age_group.value

                if issuer.native_language not in ['', None]:
                    issuer_data['native_language'] = issuer.native_language

            elif type(issuer) == issuers.Artist:
                if issuer.name not in ['', None]:
                    issuer_data['name'] = issuer.name

            data[issuer.idx] = issuer_data

        jsonfile.write_json_to_file(file_path, data)

    @staticmethod
    def write_utterances(utterance_path, corpus):
        utterance_records = {}

        for utterance in corpus.utterances.values():
            track_idx = utterance.track.idx
            start = utterance.start
            end = utterance.end

            if end == float('inf'):
                end = -1

            utterance_records[utterance.idx] = [track_idx, start, end]

        textfile.write_separated_lines(utterance_path, utterance_records, separator=' ', sort_by_column=0)

    @staticmethod
    def write_utt_to_issuer_mapping(utt_issuer_path, corpus):
        utt_issuer_records = {}

        for utterance in corpus.utterances.values():
            if utterance.issuer is not None:
                utt_issuer_records[utterance.idx] = utterance.issuer.idx

        textfile.write_separated_lines(utt_issuer_path, utt_issuer_records, separator=' ', sort_by_column=0)

    @staticmethod
    def write_labels(path, corpus):
        records = collections.defaultdict(list)

        for utterance in corpus.utterances.values():
            for label_list_idx, label_list in utterance.label_lists.items():
                utt_records = []
                for l in label_list:
                    start = l.start
                    end = l.end

                    if end == float('inf'):
                        end = -1

                    if len(l.meta) > 0:
                        value = '{} [{}]'.format(l.value, json.dumps(l.meta, sort_keys=True))
                        utt_records.append((utterance.idx, start, end, value))
                    else:
                        utt_records.append((utterance.idx, start, end, l.value))

                records[label_list_idx].extend(utt_records)

        for label_list_idx, label_list_records in records.items():
            file_path = os.path.join(path, '{}_{}.txt'.format(LABEL_FILE_PREFIX, label_list_idx))
            textfile.write_separated_lines(file_path, label_list_records, separator=' ')

    @staticmethod
    def write_feature_containers(container_path, corpus):
        feat_records = [(idx, container.path) for idx, container in corpus.feature_containers.items()]
        textfile.write_separated_lines(container_path, feat_records, separator=' ')

    @staticmethod
    def write_subviews(path, corpus):
        for name, sv in corpus.subviews.items():
            sv_path = os.path.join(path, '{}_{}.txt'.format(SUBVIEW_FILE_PREFIX, name))
            with open(sv_path, 'w') as f:
                f.write(sv.serialize())



"""
Wrapping opening function of audioread library.
This is used to cache the available backends.
If backend evaluation is done on every call it is very inefficient.
"""

import audioread

available_backends = audioread.available_backends()


def audio_open(path):
    return audioread.audio_open(path, backends=available_backends)



import os
import re

import audiomate
from audiomate import annotations
from . import base
from . import downloader

DATA_URL = 'http://asi.insa-rouen.fr/enseignants/~arakoto/data_rouen.zip'

LABEL_PATTERN = r'(.*?)(\d+)'


class RouenDownloader(downloader.ArchiveDownloader):
    """
    Downloader for the LITIS Rouen Audio scene dataset.
    """

    def __init__(self):
        super(RouenDownloader, self).__init__(
            DATA_URL,
            move_files_up=True
        )

    @classmethod
    def type(cls):
        return 'rouen'


class RouenReader(base.CorpusReader):
    """
    Reader for the LITIS Rouen Audio scene dataset.

    .. seealso::

       `Rouen <https://sites.google.com/site/alainrakotomamonjy/home/audio-scene>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'rouen'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)
        regex = re.compile(LABEL_PATTERN)

        for file_name in os.listdir(path):
            base_name, ext = os.path.splitext(file_name)

            if ext == '.wav':
                file_path = os.path.join(path, file_name)

                match = regex.match(base_name)
                label = match.group(1)

                corpus.new_file(file_path, base_name)
                utt = corpus.new_utterance(base_name, base_name)
                ll = annotations.LabelList.create_single(
                    label,
                    idx=audiomate.corpus.LL_SOUND_CLASS
                )
                utt.set_label_list(ll)

        return corpus



import os
import tempfile
import pytest

from audiomate import corpus
from audiomate.corpus import io
from audiomate.utils import textfile

from tests import resources


@pytest.fixture()
def writer():
    return io.MozillaDeepSpeechWriter()


@pytest.fixture()
def path():
    return tempfile.mkdtemp()


class TestMozillaDeepSpeechWriter:

    def test_save_all(self, writer, tmpdir):
        ds = resources.create_dataset()
        writer.save(ds, tmpdir.strpath)

        all_path = os.path.join(tmpdir.strpath, 'all.csv')

        assert os.path.isfile(all_path)

        records = textfile.read_separated_lines(all_path, separator=',')

        assert len(records) == 6

        # HEADER
        assert len(records[0]) == 3
        assert records[0][1] == 'wav_filesize'
        assert records[0][2] == 'transcript'

        # DATA RECORDS
        utts = {r[0]: (r[1], r[2]) for r in records[1:]}

        path = ds.utterances['utt-1'].track.path
        assert len(utts[path]) == 2
        assert utts[path][0] == '83090'
        assert utts[path][1] == ds.utterances['utt-1'].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        path = ds.utterances['utt-2'].track.path
        assert len(utts[path]) == 2
        assert utts[path][0] == '83090'
        assert utts[path][1] == ds.utterances['utt-2'].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        path = os.path.join(tmpdir.strpath, 'audio', 'utt-3.wav')
        assert len(utts[path]) == 2
        assert utts[path][0] == '48044'
        assert utts[path][1] == ds.utterances['utt-3'].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        path = os.path.join(tmpdir.strpath, 'audio', 'utt-4.wav')
        assert len(utts[path]) == 2
        assert utts[path][0] == '32044'
        assert utts[path][1] == ds.utterances['utt-4'].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        path = ds.utterances['utt-5'].track.path
        assert len(utts[path]) == 2
        assert utts[path][0] == '83090'
        assert utts[path][1] == ds.utterances['utt-5'].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

    def test_save_subset_train(self, writer, tmpdir):
        ds = resources.create_dataset()
        writer.save(ds, tmpdir.strpath)

        all_path = os.path.join(tmpdir.strpath, 'train.csv')

        assert os.path.isfile(all_path)

        records = textfile.read_separated_lines(all_path, separator=',')

        assert len(records) == 4

        # HEADER
        assert len(records[0]) == 3
        assert records[0][1] == 'wav_filesize'
        assert records[0][2] == 'transcript'

        # DATA RECORDS
        utts = {r[0]: (r[1], r[2]) for r in records[1:]}

        path = ds.utterances['utt-1'].track.path
        assert len(utts[path]) == 2
        assert utts[path][0] == '83090'
        assert utts[path][1] == ds.utterances['utt-1'].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        path = ds.utterances['utt-2'].track.path
        assert len(utts[path]) == 2
        assert utts[path][0] == '83090'
        assert utts[path][1] == ds.utterances['utt-2'].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        path = os.path.join(tmpdir.strpath, 'audio', 'utt-3.wav')
        assert len(utts[path]) == 2
        assert utts[path][0] == '48044'
        assert utts[path][1] == ds.utterances['utt-3'].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

    def test_save_subset_dev(self, writer, tmpdir):
        ds = resources.create_dataset()
        writer.save(ds, tmpdir.strpath)

        all_path = os.path.join(tmpdir.strpath, 'dev.csv')

        assert os.path.isfile(all_path)

        records = textfile.read_separated_lines(all_path, separator=',')

        assert len(records) == 3

        # HEADER
        assert len(records[0]) == 3
        assert records[0][1] == 'wav_filesize'
        assert records[0][2] == 'transcript'

        # DATA RECORDS
        utts = {r[0]: (r[1], r[2]) for r in records[1:]}

        path = os.path.join(tmpdir.strpath, 'audio', 'utt-4.wav')
        assert len(utts[path]) == 2
        assert utts[path][0] == '32044'
        assert utts[path][1] == ds.utterances['utt-4'].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        path = ds.utterances['utt-5'].track.path
        assert len(utts[path]) == 2
        assert utts[path][0] == '83090'
        assert utts[path][1] == ds.utterances['utt-5'].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value



import copy

from .issuer import Issuer


class Artist(Issuer):
    """
    The artist is the person/group who have produced a musical segment
    in a utterance.

    Args:
        idx (str): An unique identifier for this speaker within a dataset.
        name (str): The name of the artist/band/...
        info (dict): Any additional info for this speaker as dict.

    Attributes:
        Issuer.utterances (list): List of utterances that this issuer owns.
    """

    __slots__ = ['name']

    def __init__(self, idx, name, info={}):
        super(Artist, self).__init__(idx, info=info)

        self.name = name

    def __str__(self):
        return 'Artist(idx={0}, info={1})'.format(self.idx, self.info)

    def __copy__(self):
        # self.utterances is ignored intentionally
        # only a "weak-ref" when added to a corpus

        cp = Artist(
            self.idx,
            self.name,
            info=self.info
        )

        return cp

    def __deepcopy__(self, memo):
        # self.utterances is ignored intentionally
        # only a "weak-ref" when added to a corpus

        cp = Artist(
            self.idx,
            self.name,
            info=copy.deepcopy(self.info, memo)
        )

        return cp



import abc
import enum
import os
import tarfile
import zipfile

from audiomate.utils import download
from audiomate.utils import files
from . import base


class ArkType(enum.Enum):
    """
    Enum defining different types of archives.
    """
    ZIP = 1
    TAR = 2
    AUTO = 3


class ArchiveDownloader(base.CorpusDownloader, abc.ABC):
    """
    Convenience base class for a downloader of a corpus,
    that consists of a single archive.

    Args:
        url (str): URL, from where to download the archive.
        ark_type (ArkType): The type of the archive.
                            If ``AUTO`` it tries to find the type
                            automatically.
        move_files_up (bool): If ``True`` moves all files/folders
                              from subfolders to the root-folder.
    """

    def __init__(self, url, ark_type=ArkType.AUTO, move_files_up=False):
        self.url = url
        self.ark_type = ark_type
        self.move_files_up = move_files_up

    def _download(self, target_path):
        os.makedirs(target_path, exist_ok=True)
        tmp_file = os.path.join(target_path, 'tmp_ark')

        download.download_file(self.url, tmp_file)
        self._extract_file(tmp_file, target_path)

        if self.move_files_up:
            files.move_all_files_from_subfolders_to_top(
                target_path,
                delete_subfolders=True
            )

        os.remove(tmp_file)

    def _extract_file(self,  file_path, target_folder):
        ark_type = self.ark_type

        if self.ark_type == ArkType.AUTO:
            if tarfile.is_tarfile(file_path):
                ark_type = ArkType.TAR
            elif zipfile.is_zipfile(file_path):
                ark_type = ArkType.ZIP

        if ark_type == ArkType.TAR:
            download.extract_tar(file_path, target_folder)
        elif ark_type == ArkType.ZIP:
            download.extract_zip(file_path, target_folder)
        else:
            raise ValueError(
                'Unrecognized archive type (Only zip/tar supported)!'
            )



"""
This module contains functionality for creating any kind of subsets from a corpus.
A subset of a corpus is represented with a :py:class:`Subview`.
The data contained in a subview is defined by one or more :py:class:`FilterCriterion`.

For creating subviews there are additional classes.
:py:class:`Splitter` can be used to divide a corpus into subsets according to given proportions.
:py:class:`SubsetGenerator` can be used to create subset with given settings.


Subview
-------
.. autoclass:: Subview
   :members:
   :inherited-members:


Filter
------
.. autoclass:: FilterCriterion
   :members:
   :inherited-members:


MatchingUtteranceIdxFilter
^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: MatchingUtteranceIdxFilter


MatchingLabelFilter
^^^^^^^^^^^^^^^^^^^
.. autoclass:: MatchingLabelFilter


Splitter
--------
.. autoclass:: Splitter
   :members:
   :inherited-members:


SubsetGenerator
---------------

.. autoclass:: SubsetGenerator
   :members:
   :inherited-members:


Utils
-----

.. automodule:: audiomate.corpus.subset.utils
   :members:

"""

from .subview import FilterCriterion  # noqa: F401
from .subview import MatchingUtteranceIdxFilter  # noqa: F401
from .subview import MatchingLabelFilter  # noqa: F401

from .subview import Subview  # noqa: F401

from .splitting import Splitter  # noqa: F401
from .selection import SubsetGenerator  # noqa: F401



import os
import re
import tarfile
import shutil

import requests

import audiomate
from . import base
from audiomate import annotations
from audiomate import issuers
from audiomate.utils import textfile

DOWNLOAD_URL = {
    'de': 'http://www.repository.voxforge1.org/downloads/de/Trunk/Audio/Main/16kHz_16bit/',
    'en': 'http://www.repository.voxforge1.org/downloads/SpeechCorpus/Trunk/Audio/Main/16kHz_16bit/'
}

BAD_UTTERANCES = [
    'anonymous-20081112-ssu-de10-024',
    'rwunsch-20090706-any-de2-26',
    'john_doe-20160503-gaj-de7-082'
]


class VoxforgeDownloader(base.CorpusDownloader):
    """
    Downloader for audio files from http://www.voxforge.org/.
    All .tgz files that are linked from the given url are downloaded and extracted.

    Args:
        lang (str): If no URL is given the predefined URL's for the given language is used, if one is defined.
        url (str): The url to check for available .tgz files.
    """

    def __init__(self, lang='de', url=None):
        self.url = url

        if url is None:
            if lang in DOWNLOAD_URL.keys():
                self.url = DOWNLOAD_URL[lang]
            else:
                raise ValueError('There is no voxforge URL present for language {}!'.format(lang))

    @classmethod
    def type(cls):
        return 'voxforge'

    def _download(self, target_path):
        temp_folder = os.path.join(target_path, 'download')
        os.makedirs(temp_folder, exist_ok=True)

        available = VoxforgeDownloader.available_files(self.url)
        downloaded = VoxforgeDownloader.download_files(available, temp_folder)
        VoxforgeDownloader.extract_files(downloaded, target_path)

        shutil.rmtree(temp_folder)

    @staticmethod
    def available_files(url):
        """ Extract and return urls for all available .tgz files. """
        req = requests.get(url)

        if req.status_code != 200:
            raise base.FailedDownloadException('Failed to download data (status {}) from {}!'.format(req.status_code,
                                                                                                     url))

        page_content = req.text
        link_pattern = re.compile(r'<a href="(.*?)">(.*?)</a>')
        available_files = []

        for match in link_pattern.findall(page_content):
            if match[0].endswith('.tgz'):
                available_files.append(os.path.join(url, match[0]))

        return available_files

    @staticmethod
    def download_files(file_urls, target_path):
        """ Download all files and store to the given path. """
        os.makedirs(target_path, exist_ok=True)
        downloaded_files = []

        for file_url in file_urls:
            req = requests.get(file_url)

            if req.status_code != 200:
                raise base.FailedDownloadException('Failed to download file {} (status {})!'.format(req.status_code,
                                                                                                    file_url))

            file_name = os.path.basename(file_url)
            target_file_path = os.path.join(target_path, file_name)

            with open(target_file_path, 'wb') as f:
                f.write(req.content)

            downloaded_files.append(target_file_path)

        return downloaded_files

    @staticmethod
    def extract_files(file_paths, target_path):
        """ Unpack all files to the given path. """
        os.makedirs(target_path, exist_ok=True)
        extracted = []

        for file_path in file_paths:
            with tarfile.open(file_path, 'r') as archive:
                archive.extractall(target_path)

            file_name = os.path.splitext(os.path.basename(file_path))[0]
            extracted.append(os.path.join(target_path, file_name))

        return extracted


class VoxforgeReader(base.CorpusReader):
    """
    Reader for collections of voxforge audio data. The reader expects extracted .tgz files in the given folder.

    .. seealso::

       `<http://www.voxforge.org/>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'voxforge'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        for dir_path in VoxforgeReader.data_folders(path):
            item = os.path.basename(dir_path)
            etc_folder = os.path.join(dir_path, 'etc')
            wav_folder = os.path.join(dir_path, 'wav')
            readme_path = os.path.join(etc_folder, 'README')

            # LOAD ISSUER
            issuer = VoxforgeReader.parse_speaker_info(readme_path)

            if issuer.idx is None or issuer.idx == 'anonymous':
                issuer.idx = item

            # LOAD TRANSCRIPTIONS
            prompts, prompts_orig = VoxforgeReader.parse_prompts(etc_folder)

            # LOAD FILES/UTTS
            for file_name in os.listdir(wav_folder):
                wav_path = os.path.join(wav_folder, file_name)
                basename, ext = os.path.splitext(file_name)
                idx = '{}-{}'.format(item, basename)

                is_valid_wav = os.path.isfile(wav_path) and ext == '.wav' and idx not in BAD_UTTERANCES
                has_transcription = basename in prompts.keys()

                if is_valid_wav and has_transcription:
                    if issuer.idx not in corpus.issuers.keys():
                        corpus.import_issuers([issuer])

                    corpus.new_file(wav_path, idx)
                    utt = corpus.new_utterance(idx, idx, issuer.idx)
                    utt.set_label_list(annotations.LabelList.create_single(prompts[basename],
                                                                           idx=audiomate.corpus.LL_WORD_TRANSCRIPT))

                    if basename in prompts_orig.keys():
                        raw = annotations.LabelList.create_single(prompts_orig[basename],
                                                                  idx=audiomate.corpus.LL_WORD_TRANSCRIPT_RAW)
                        utt.set_label_list(raw)

        return corpus

    @staticmethod
    def data_folders(path):
        """ Generator which yields a list of valid data directories (corresponds to the content of one .tgz). """
        for item in os.listdir(path):
            dir_path = os.path.join(path, item)
            wav_folder = os.path.join(dir_path, 'wav')

            if os.path.isdir(dir_path) and os.path.isdir(wav_folder):
                yield dir_path

    @staticmethod
    def parse_speaker_info(readme_path):
        """ Parse speaker info and return tuple (idx, gender). """
        idx = None
        gender = issuers.Gender.UNKNOWN
        age_group = issuers.AgeGroup.UNKNOWN
        native_lang = None

        with open(readme_path, 'r', errors='ignore') as f:
            for raw_line in f:
                line = raw_line.strip()

                if line is not None and line is not '':
                    line = line.rstrip(';.')
                    parts = line.split(':', maxsplit=1)

                    if len(parts) > 1:
                        key = parts[0].strip().lower()
                        value = parts[1].strip()

                        if key == 'user name':
                            idx = value

                        value = value.lower()

                        if key == 'gender':
                            if value in ['männlich', 'male', 'mnnlich']:
                                gender = issuers.Gender.MALE
                            elif value in ['weiblich', 'female', '[female]']:
                                gender = issuers.Gender.FEMALE

                        if key == 'age range':
                            if value in ['erwachsener', 'adult', '[adult]', '[erwachsener]']:
                                age_group = issuers.AgeGroup.ADULT
                            elif value in ['senior', '[senior']:
                                age_group = issuers.AgeGroup.SENIOR
                            elif value in ['youth', 'jugendlicher', '[youth]', '[jugendlicher]']:
                                age_group = issuers.AgeGroup.YOUTH
                            elif value in ['kind', 'child']:
                                age_group = issuers.AgeGroup.CHILD

                        if key == 'language':
                            if value in ['de', 'ger', 'deu', '[de]']:
                                native_lang = 'deu'
                            elif value in ['en', 'eng', '[en]']:
                                native_lang = 'eng'

        return issuers.Speaker(idx, gender=gender, age_group=age_group, native_language=native_lang)

    @staticmethod
    def parse_prompts(etc_folder):
        """ Read prompts and prompts-orignal and return as dictionary (id as key). """
        prompts_path = os.path.join(etc_folder, 'PROMPTS')
        prompts_orig_path = os.path.join(etc_folder, 'prompts-original')

        prompts = textfile.read_key_value_lines(prompts_path, separator=' ')
        prompts_orig = textfile.read_key_value_lines(prompts_orig_path, separator=' ')

        prompts_key_fixed = {}

        for k, v in prompts.items():
            parts = k.split('/')
            key = k

            if len(parts) > 1:
                key = parts[-1]

            prompts_key_fixed[key] = v

        prompts = prompts_key_fixed

        return prompts, prompts_orig



import numpy as np

from audiomate import tracks
from audiomate import annotations
from audiomate.utils import units
from audiomate import encoding

from tests import resources


class TestFrameHotEncoder:

    def test_encode_full_utterance(self):
        ds = resources.create_multi_label_corpus()
        enc = encoding.FrameHotEncoder(['music', 'speech', 'noise'],
                                       'default',
                                       frame_settings=units.FrameSettings(32000, 16000),
                                       sr=16000)

        actual = enc.encode_utterance(ds.utterances['utt-6'])
        expected = np.array([
            [1, 0, 0],
            [1, 0, 0],
            [1, 0, 0],
            [1, 0, 0],
            [1, 1, 0],
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0],
            [1, 0, 0],
            [1, 0, 0],
        ]).astype(np.float32)

        assert np.array_equal(expected, actual)

    def test_encode_label_ends_at_utterance_end(self):
        track = tracks.FileTrack('file1', resources.sample_wav_file('med_len.wav'))
        utt = tracks.Utterance('utt1', track, start=3, end=14)
        ll = annotations.LabelList(labels=[
            annotations.Label('speech', 0, 4),
            annotations.Label('music', 4, 9),
            annotations.Label('speech', 9, float('inf')),
        ])
        utt.set_label_list(ll)

        enc = encoding.FrameHotEncoder(['music', 'speech', 'noise'],
                                       'default',
                                       frame_settings=units.FrameSettings(32000, 16000),
                                       sr=16000)

        actual = enc.encode_utterance(utt)
        expected = np.array([
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0],
            [1, 1, 0],
            [1, 0, 0],
            [1, 0, 0],
            [1, 0, 0],
            [1, 0, 0],
            [1, 1, 0],
            [0, 1, 0],
        ]).astype(np.float32)

        assert np.array_equal(expected, actual)

    def test_encode_label_ends_at_track_end(self):
        track = tracks.FileTrack('file1', resources.sample_wav_file('med_len.wav'))
        utt = tracks.Utterance('utt1', track, start=3, end=float('inf'))
        ll = annotations.LabelList(labels=[
            annotations.Label('speech', 0, 4),
            annotations.Label('music', 4, 9),
            annotations.Label('speech', 9, float('inf')),
        ])
        utt.set_label_list(ll)

        enc = encoding.FrameHotEncoder(['music', 'speech', 'noise'],
                                       'default',
                                       frame_settings=units.FrameSettings(32000, 16000),
                                       sr=16000)

        actual = enc.encode_utterance(utt)
        expected = np.array([
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0],
            [1, 1, 0],
            [1, 0, 0],
            [1, 0, 0],
            [1, 0, 0],
            [1, 0, 0],
            [1, 1, 0],
            [0, 1, 0],
            [0, 1, 0],
            [0, 1, 0],
        ]).astype(np.float32)

        for r in actual:
            print(r)

        assert np.array_equal(expected, actual)


class TestFrameOrdinalEncoder:

    def test_encode_utterance(self):
        ds = resources.create_multi_label_corpus()
        enc = encoding.FrameOrdinalEncoder(['music', 'speech', 'noise'],
                                           'default',
                                           frame_settings=units.FrameSettings(32000, 16000),
                                           sr=16000)

        actual = enc.encode_utterance(ds.utterances['utt-6'])
        expected = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0]).astype(np.int)

        assert np.array_equal(expected, actual)

    def test_encode_utterance_takes_larger_label(self):
        file = tracks.FileTrack('file-idx', resources.sample_wav_file('wav_1.wav'))
        utt = tracks.Utterance('utt-idx', file, start=0, end=8)
        ll = annotations.LabelList(labels=[
            annotations.Label('music', 0, 4.5),
            annotations.Label('speech', 4.5, 8)
        ])
        utt.set_label_list(ll)

        enc = encoding.FrameOrdinalEncoder(['music', 'speech', 'noise'],
                                           'default',
                                           frame_settings=units.FrameSettings(32000, 16000),
                                           sr=16000)

        actual = enc.encode_utterance(utt)
        expected = np.array([0, 0, 0, 0, 1, 1, 1]).astype(np.int)

        assert np.array_equal(expected, actual)

    def test_encode_utterance_takes_lower_index_first(self):
        file = tracks.FileTrack('file-idx', resources.sample_wav_file('wav_1.wav'))
        utt = tracks.Utterance('utt-idx', file, start=0, end=5)
        ll = annotations.LabelList(labels=[
            annotations.Label('music', 0, 3),
            annotations.Label('speech', 3, 5)
        ])
        utt.set_label_list(ll)

        enc = encoding.FrameOrdinalEncoder(['speech', 'music', 'noise'],
                                           'default',
                                           frame_settings=units.FrameSettings(32000, 16000),
                                           sr=16000)

        actual = enc.encode_utterance(utt)
        expected = np.array([1, 1, 0, 0]).astype(np.int)

        assert np.array_equal(expected, actual)



import abc

from audiomate import containers


class Encoder(metaclass=abc.ABCMeta):
    """
    Base class for an encoder. The goal of an encoder is to extract encoded targets for an utterance.
    The base class provides functionality to perform encoding for a full corpus.
    A concrete encoder just has to provide the method to encode a single utterance via ``encode_utterance``.

    For example for training a frame-classifier, an encoder extracts one-hot encoded vectors from a label-list.
    """

    def encode_corpus(self, corpus, output_path):
        """
        Encode all utterances of the given corpus and store them in a :class:`audiomate.container.Container`.

        Args:
            corpus (Corpus): The corpus to process.
            output_path (str): The path to store the container with the encoded data.

        Returns:
            Container: The container with the encoded data.
        """

        out_container = containers.Container(output_path)
        out_container.open()

        for utterance in corpus.utterances.values():
            data = self.encode_utterance(utterance, corpus=corpus)
            out_container.set(utterance.idx, data)

        out_container.close()
        return out_container

    @abc.abstractmethod
    def encode_utterance(self, utterance, corpus=None):
        """
        Encode the given utterance.

        Args:
            utterance (Utterance): The utterance to encode.
            corpus (Corpus): The corpus the utterance is from.

        Returns:
            np.ndarray: Encoded data.
        """
        pass



import numpy as np

from audiomate.processing import pipeline


class TestMeanVarianceNorm:

    def test_compute(self):
        frame = np.random.random_sample(5)
        mean = float(np.mean(frame))
        var = float(np.var(frame))

        norm = pipeline.MeanVarianceNorm(mean, var)
        output = norm.process_frames(frame, 4)
        expected = (frame - mean) / np.std(frame)

        assert np.array_equal(output, expected)



import copy
import enum

from .issuer import Issuer


class Gender(enum.Enum):
    UNKNOWN = 'unknown'
    MALE = 'male'
    FEMALE = 'female'


class AgeGroup(enum.Enum):
    UNKNOWN = 'unknown'
    CHILD = 'child'
    YOUTH = 'youth'
    ADULT = 'adult'
    SENIOR = 'senior'


class Speaker(Issuer):
    """
    The speaker is the person who spoke in a utterance.

    Args:
        idx (str): An unique identifier for this speaker within a dataset.
        info (dict): Any additional info for this speaker as dict.
        age_group (AgeGroup): The age-group of the speaker (child, adult, ...)
        native_language (str): The native language of the speaker. (ISO 639-3)

    Attributes:
        Issuer.utterances (list): List of utterances that this issuer owns.
    """

    __slots__ = ['gender', 'age_group', 'native_language']

    def __init__(self, idx, gender=Gender.UNKNOWN, age_group=AgeGroup.UNKNOWN,
                 native_language=None, info={}):
        super(Speaker, self).__init__(idx, info=info)

        self.gender = gender
        self.age_group = age_group
        self.native_language = native_language

    def __str__(self):
        return 'Speaker(idx={0}, info={1})'.format(self.idx, self.info)

    def __copy__(self):
        # self.utterances is ignored intentionally
        # only a "weak-ref" when added to a corpus

        cp = Speaker(
            self.idx,
            gender=self.gender,
            age_group=self.age_group,
            native_language=self.native_language,
            info=self.info
        )

        return cp

    def __deepcopy__(self, memo):
        # self.utterances is ignored intentionally
        # only a "weak-ref" when added to a corpus

        cp = Speaker(
            self.idx,
            gender=self.gender,
            age_group=self.age_group,
            native_language=self.native_language,
            info=copy.deepcopy(self.info, memo)
        )

        return cp



import os

import numpy as np

from audiomate import containers
import pytest

from tests import resources


@pytest.fixture()
def sample_container():
    container_path = resources.get_resource_path(
        ['sample_files', 'feat_container']
    )
    sample_container = containers.Container(container_path)
    sample_container.open()
    yield sample_container
    sample_container.close()


class TestContainer:

    def test_is_open(self, sample_container):
        assert sample_container.is_open()

    def test_is_open_returns_false(self, sample_container):
        sample_container.close()
        assert not sample_container.is_open()

    def test_keys(self, sample_container):
        assert sample_container.keys() == ['utt-1', 'utt-2', 'utt-3']

    def test_get(self, sample_container):
        assert sample_container.get('utt-1', mem_map=False).shape == (20, 5)

    def test_remove(self, sample_container):
        sample_container.set('some-key', np.arange(20))
        assert sample_container.keys() == ['some-key', 'utt-1', 'utt-2', 'utt-3']

        sample_container.remove('some-key')
        assert sample_container.keys() == ['utt-1', 'utt-2', 'utt-3']

    def test_append(self, tmpdir):
        path = os.path.join(tmpdir.strpath, 'container')
        tmp_container = containers.Container(path)
        tmp_container.open()

        data = np.arange(100).reshape(20, 5)

        tmp_container.append('utt-1', data[:8])
        tmp_container.append('utt-1', data[8:])

        res = tmp_container.get('utt-1', mem_map=False)

        assert np.array_equal(data, res)

        tmp_container.close()

    def test_append_with_different_dimension_raises_error(self, tmpdir):
        path = os.path.join(tmpdir.strpath, 'container')
        tmp_container = containers.Container(path)
        tmp_container.open()

        tmp_container.append('utt-1', np.arange(20).reshape(5, 2, 2))

        with pytest.raises(ValueError):
            tmp_container.append('utt-1', np.arange(42).reshape(7, 2, 3))

        tmp_container.close()



import os
import numpy as np

from audiomate import containers
from audiomate import tracks

import pytest

from tests import resources


@pytest.fixture
def sample_container():
    container_path = resources.get_resource_path(
        ['sample_files', 'audio_container']
    )
    sample_container = containers.AudioContainer(container_path)
    sample_container.open()
    yield sample_container
    sample_container.close()


class TestContainerTrack:

    def test_sampling_rate(self, sample_container):
        sample_track = tracks.ContainerTrack('track1', sample_container)
        assert sample_track.sampling_rate == 16000

    def test_num_samples(self, sample_container):
        sample_track = tracks.ContainerTrack('track1', sample_container)
        assert sample_track.num_samples == 10

    def test_duration(self, sample_container):
        sample_track = tracks.ContainerTrack('track1', sample_container)
        assert sample_track.duration == pytest.approx(10 / 16000)

    def test_read_samples(self, sample_container):
        sample_track = tracks.ContainerTrack('track1', sample_container)
        samples = sample_track.read_samples()

        assert samples.dtype == np.float32
        assert np.allclose(
            samples,
            np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]),
            atol=1.e-4
        )

    def test_read_samples_with_offset(self, sample_container):
        sample_track = tracks.ContainerTrack('track1', sample_container)
        samples = sample_track.read_samples(offset=5/16000)

        assert samples.dtype == np.float32
        assert np.allclose(
            samples,
            np.array([0.6, 0.7, 0.8, 0.9, 1.0]),
            atol=1.e-4
        )

    def test_read_samples_with_duration(self, sample_container):
        sample_track = tracks.ContainerTrack('track1', sample_container)
        samples = sample_track.read_samples(duration=5/16000)

        assert samples.dtype == np.float32
        assert np.allclose(
            samples,
            np.array([0.1, 0.2, 0.3, 0.4, 0.5]),
            atol=1.e-4
        )

    def test_read_samples_with_resampling(self, sample_container):
        sample_track = tracks.ContainerTrack('track1', sample_container)
        samples = sample_track.read_samples(sr=8000)

        assert samples.dtype == np.float32
        assert samples.shape[0] == 5

    def test_read_frames(self, tmpdir):
        cont_path = os.path.join(tmpdir.strpath, 'audio.hdf5')
        cont = containers.AudioContainer(cont_path)
        cont.open()

        content = np.random.random(10044)
        cont.set('track', content, 16000)
        track = tracks.ContainerTrack('some_idx', cont, 'track')

        data = list(track.read_frames(frame_size=400, hop_size=160))
        frames = np.array([x[0] for x in data])
        last = [x[1] for x in data]

        assert frames.shape == (62, 400)
        assert frames.dtype == np.float32
        assert np.allclose(frames[0], content[:400], atol=0.0001)
        expect = np.pad(content[9760:], (0, 116), mode='constant')
        assert np.allclose(frames[61], expect, atol=0.0001)

        assert last[:-1] == [False] * (len(data) - 1)
        assert last[-1]

        cont.close()

    def test_read_frames_matches_length(self, tmpdir):
        cont_path = os.path.join(tmpdir.strpath, 'audio.hdf5')
        cont = containers.AudioContainer(cont_path)
        cont.open()

        content = np.random.random(7)
        cont.set('track', content, 16000)
        track = tracks.ContainerTrack('some_idx', cont, 'track')

        data = list(track.read_frames(frame_size=2, hop_size=1))
        frames = np.array([x[0] for x in data])
        last = [x[1] for x in data]

        assert frames.shape == (6, 2)
        assert frames.dtype == np.float32

        assert last[:-1] == [False] * (len(data) - 1)
        assert last[-1]

        cont.close()



from . import base

import audioread


class TrackReadValidator(base.Validator):
    """
    Check if the track can be opened and read.
    """

    def name(self):
        return 'Track-Read'

    def validate(self, corpus):
        """
        Perform the validation on the given corpus.

        Args:
            corpus (Corpus): The corpus to test/validate.

        Returns:
            InvalidItemsResult: Validation result.
        """

        invalid_tracks = {}

        for track in corpus.tracks.values():
            try:
                track.duration
            except EOFError:
                invalid_tracks[track.idx] = 'EOFError'
            except audioread.NoBackendError:
                invalid_tracks[track.idx] = 'NoBackendError'

        passed = len(invalid_tracks) <= 0
        return base.InvalidItemsResult(
            passed,
            invalid_tracks,
            item_name='Tracks',
            name=self.name()
        )



import pytest

from audiomate import annotations
from audiomate.corpus import subset
from tests import resources


@pytest.fixture()
def corpus():
    return resources.create_single_label_corpus()


@pytest.fixture()
def corpus_with_more_labels():
    """
    Corpus with an extra label-list.
    """
    corpus = resources.create_single_label_corpus()

    corpus.utterances['utt-1'].set_label_list(annotations.LabelList(idx='radio', labels=[
        annotations.Label('alpha')
    ]))

    corpus.utterances['utt-2'].set_label_list(annotations.LabelList(idx='radio', labels=[
        annotations.Label('alpha')
    ]))

    corpus.utterances['utt-3'].set_label_list(annotations.LabelList(idx='radio', labels=[
        annotations.Label('beta')
    ]))

    corpus.utterances['utt-4'].set_label_list(annotations.LabelList(idx='radio', labels=[
        annotations.Label('beta')
    ]))

    corpus.utterances['utt-5'].set_label_list(annotations.LabelList(idx='radio', labels=[
        annotations.Label('beta')
    ]))

    corpus.utterances['utt-6'].set_label_list(annotations.LabelList(idx='radio', labels=[
        annotations.Label('beta')
    ]))

    corpus.utterances['utt-7'].set_label_list(annotations.LabelList(idx='radio', labels=[
        annotations.Label('beta')
    ]))

    corpus.utterances['utt-8'].set_label_list(annotations.LabelList(idx='radio', labels=[
        annotations.Label('beta')
    ]))

    return corpus


class TestSubsetGenerator:

    def test_random_subset(self, corpus):
        g = subset.SubsetGenerator(corpus, random_seed=20)
        sv = g.random_subset(0.5)

        assert sv.num_utterances == 4

    def test_random_subset_balanced(self, corpus):
        g = subset.SubsetGenerator(corpus, random_seed=20)
        sv = g.random_subset(0.5, balance_labels=True)

        label_count = sv.label_count()

        assert sv.num_utterances == 4
        assert label_count['music'] / label_count['speech'] == pytest.approx(1.0, abs=0.5)

    def test_random_subset_balanced_on_single_label_list(self, corpus_with_more_labels):
        g = subset.SubsetGenerator(corpus_with_more_labels, random_seed=20)
        sv = g.random_subset(0.5, balance_labels=True, label_list_ids=['radio'])

        assert sv.num_utterances == 4

    def test_random_subset_by_duration(self, corpus):
        g = subset.SubsetGenerator(corpus, random_seed=20)
        sv = g.random_subset_by_duration(0.5)

        assert sv.total_duration == pytest.approx(corpus.total_duration * 0.5, abs=5)

    def test_random_subset_by_duration_balanced(self, corpus):
        g = subset.SubsetGenerator(corpus, random_seed=20)
        sv = g.random_subset_by_duration(0.5, balance_labels=True)

        label_durations = sv.label_durations()

        assert sv.total_duration == pytest.approx(corpus.total_duration * 0.5, abs=15)
        assert label_durations['music'] / label_durations['speech'] == pytest.approx(1.0, abs=0.5)

    def test_random_subset_by_duration_balanced_on_single_label_list(self, corpus_with_more_labels):
        g = subset.SubsetGenerator(corpus_with_more_labels, random_seed=20)
        sv = g.random_subset_by_duration(0.5, balance_labels=True, label_list_ids=['radio'])

        assert sv.total_duration == pytest.approx(corpus_with_more_labels.total_duration * 0.5, abs=15)

    def test_random_subsets(self, corpus):
        g = subset.SubsetGenerator(corpus, random_seed=20)
        svs = g.random_subsets([0.6, 0.4, 0.2])

        assert svs.keys() == {0.6, 0.4, 0.2}

        assert svs[0.6].num_utterances == 5
        assert svs[0.4].num_utterances == 3
        assert svs[0.2].num_utterances == 2

        assert set(svs[0.4].utterances.keys()).issubset(set(svs[0.6].utterances.keys()))
        assert set(svs[0.2].utterances.keys()).issubset(set(svs[0.4].utterances.keys()))

    def test_random_subsets_balanced_on_single_label_list(self, corpus_with_more_labels):
        g = subset.SubsetGenerator(corpus_with_more_labels, random_seed=20)
        svs = g.random_subsets([0.6, 0.4, 0.2], balance_labels=True, label_list_ids=['radio'])

        assert svs.keys() == {0.6, 0.4, 0.2}

    def test_maximal_balanced_subset(self, corpus):
        g = subset.SubsetGenerator(corpus, random_seed=20)
        sv = g.maximal_balanced_subset()

        label_count = sv.label_count()

        assert sv.num_utterances == 6
        assert label_count['music'] == 3
        assert label_count['speech'] == 3

    def test_maximal_balanced_subset_by_duration(self, corpus_with_more_labels):
        g = subset.SubsetGenerator(corpus_with_more_labels, random_seed=20)
        sv = g.maximal_balanced_subset(by_duration=True, label_list_ids=['radio'])

        label_count = sv.label_count()

        assert sv.num_utterances == 3
        assert label_count['alpha'] == 2
        assert label_count['beta'] == 1



"""
This module contains the different implementations of containers.
A container is normally used to store data of a specific type
for all instances of a corpus (e.g. mfcc-features of all utterances).

All container implementations are based on
:py:class:`audiomate.containers.Container`, which provides the basic
functionality to access a HDF5-file using h5py.
"""

from .container import Container  # noqa: F401
from .features import FeatureContainer  # noqa: F401
from .audio import AudioContainer  # noqa: F401



"""
This module contains classes to read and write corpora from the filesystem in a wide range of formats. They can also be
used to convert between formats.
"""

from .base import CorpusDownloader, CorpusReader, CorpusWriter
from .downloader import ArchiveDownloader

from .broadcast import BroadcastReader  # noqa: F401
from .default import DefaultReader, DefaultWriter  # noqa: F401
from .gtzan import GtzanDownloader, GtzanReader  # noqa: F401
from .kaldi import KaldiReader, KaldiWriter  # noqa: F401
from .musan import MusanDownloader, MusanReader  # noqa: F401
from .speech_commands import SpeechCommandsReader  # noqa: F401
from .tuda import TudaReader  # noqa: F401
from .folder import FolderReader  # noqa: F401
from .esc import ESC50Downloader, ESC50Reader  # noqa: F401
from .mozilla_deepspeech import MozillaDeepSpeechWriter  # noqa: F401
from .voxforge import VoxforgeDownloader, VoxforgeReader  # noqa: F401
from .aed import AEDReader, AEDDownloader  # noqa: F401
from .urbansound import Urbansound8kReader  # noqa: F401
from .timit import TimitReader  # noqa: F401
from .swc import SWCReader  # noqa: F401
from .free_spoken_digits import FreeSpokenDigitDownloader, FreeSpokenDigitReader  # noqa: F401
from .tatoeba import TatoebaDownloader, TatoebaReader  # noqa: F401
from .common_voice import CommonVoiceReader  # noqa: F401
from .mailabs import MailabsDownloader, MailabsReader  # noqa: F401
from .rouen import RouenDownloader, RouenReader  # noqa: F401
from .audio_mnist import AudioMNISTDownloader, AudioMNISTReader  # noqa: F401
from .fluent_speech import FluentSpeechReader  # noqa: F401
from .wav2letter import Wav2LetterWriter  # noqa: F401

__downloaders = {}
for cls in CorpusDownloader.__subclasses__():
    if cls != ArchiveDownloader:
        __downloaders[cls.type()] = cls

__readers = {}
for cls in CorpusReader.__subclasses__():
    __readers[cls.type()] = cls

__writers = {}
for cls in CorpusWriter.__subclasses__():
    __writers[cls.type()] = cls


class UnknownDownloaderException(Exception):
    pass


class UnknownReaderException(Exception):
    pass


class UnknownWriterException(Exception):
    pass


def available_downloaders():
    """
    Get a mapping of all available downloaders.

    Returns:
        dict: A dictionary with downloader classes with the name of these downloaders as key.

    Example::

        >>> available_downloaders()
        {
            "voxforge" : audiomate.corpus.io.VoxforgeDownloader
        }
    """
    return __downloaders


def available_readers():
    """
    Get a mapping of all available readers.

    Returns:
        dict: A dictionary with reader classes with the name of these readers as key.

    Example::

        >>> available_readers()
        {
            "default" : audiomate.corpus.io.DefaultReader,
            "kaldi" : audiomate.corpus.io.KaldiReader
        }
    """
    return __readers


def available_writers():
    """
    Get a mapping of all available writers.

    Returns:
        dict: A dictionary with writer classes with the name of these writers as key.

    Example::

        >>> available_writers()
        {
            "default" : audiomate.corpus.io.DefaultWriter,
            "kaldi" : audiomate.corpus.io.KaldiWriter
        }
    """
    return __writers


def create_downloader_of_type(type_name):
    """
        Create an instance of the downloader with the given name.

        Args:
            type_name: The name of a downloader.

        Returns:
            An instance of the downloader with the given type.
    """
    downloaders = available_downloaders()

    if type_name not in downloaders.keys():
        raise UnknownDownloaderException('Unknown downloader: %s' % (type_name,))

    return downloaders[type_name]()


def create_reader_of_type(type_name):
    """
        Create an instance of the reader with the given name.

        Args:
            type_name: The name of a reader.

        Returns:
            An instance of the reader with the given type.
    """
    readers = available_readers()

    if type_name not in readers.keys():
        raise UnknownReaderException('Unknown reader: %s' % (type_name,))

    return readers[type_name]()


def create_writer_of_type(type_name):
    """
        Create an instance of the writer with the given name.

        Args:
            type_name: The name of a writer.

        Returns:
            An instance of the writer with the given type.
    """
    writers = available_writers()

    if type_name not in writers.keys():
        raise UnknownWriterException('Unknown writer: %s' % (type_name,))

    return writers[type_name]()
